# This file is part of PSAMM.
#
# PSAMM is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PSAMM is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PSAMM.  If not, see <http://www.gnu.org/licenses/>.
#
# Copyright 2014-2015  Jon Lund Steffensen <jon_steffensen@uri.edu>
# Copyright 2018-2020  Jing Wang <wjingsjtu@gmail.com>

"""Utility functions."""
from __future__ import division
from future.utils import iteritems
from builtins import range

import re
from itertools import product

from psamm.formula import Atom
from psamm.expression.boolean import Expression


def id_equals(id1, id2):
    """Return True if the two IDs are considered equal."""
    return id1.lower() == id2.lower()


def name_equals(name1, name2):
    """Return True if the two names are considered equal."""
    if name1 is None or name2 is None:
        return False
    # remove special chars, translate to lower case
    pattern = r'[^a-zA-Z0-9]'
    name1 = re.sub(pattern, '', name1.lower())
    name2 = re.sub(pattern, '', name2.lower())
    # unify Coenzyme A and CoA
    pattern = r'coenzymea'
    name1 = re.sub(pattern, 'coa', name1.lower())
    name2 = re.sub(pattern, 'coa', name2.lower())
    return (name1 == name2)


def name_similar(name1, name2):
    """Return the possibility that two names are considered equal."""
    pattern = r'[()\[\]_, -]'
    if name1 is None or name2 is None:
        return 0.01
    name1 = re.sub(pattern, '', name1)
    name2 = re.sub(pattern, '', name2)
    return max(0.01,
               (1 - float(levenshtein(name1, name2)) /
                max(len(name1), len(name2))))


def formula_equals(f1, f2, charge1, charge2):
    """Return True if the two formulas are considered equal."""
    if f1 is None or f2 is None:
        return False

    def calc_formula(f, charge, neutral=False):
        formula = {}
        for e, value in iteritems(f):
            if e != Atom('H') or charge is None or not neutral:
                formula[e] = value
            else:
                formula[e] = value - charge  # No. of H in neutral stat
        return formula

    # in case some model doesn't provide charge for non-neutral formula
    # compare original formula
    formula1 = calc_formula(f1, charge1)
    formula2 = calc_formula(f2, charge2)
    # compare neutral formula
    formula1_neutral = calc_formula(f1, charge1, neutral=True)
    formula2_neutral = calc_formula(f2, charge2, neutral=True)

    return (formula1 == formula2) or (formula1_neutral == formula2_neutral)


def genes_equals(g1, g2, gene_map={}):
    """Return True if the two gene association strings are considered equal.

    Args:
        g1, g2: gene association strings
        gene_map: a dict that maps gene ids in g1 to gene ids in g2 if
                  they have different naming system
    """
    if g1 is None or g2 is None:
        return False

    e1 = Expression(g1)
    e2 = Expression(g2)

    g_list = set([gene_map.get(v.symbol, v.symbol) for v in e2.variables])
    check1 = e1.substitute(lambda v: v.symbol in g_list)

    g_list = set([gene_map.get(v.symbol, v.symbol) for v in e1.variables])
    check2 = e2.substitute(lambda v: v.symbol in g_list)

    return check1.value and check2.value


def formula_exact(f1, f2):
    """Return True if the two formulas are considered equal."""
    if f1 is None or f2 is None:
        return False

    formula1 = {e: value for e, value in iteritems(f1)}
    formula2 = {e: value for e, value in iteritems(f2)}
    return formula1 == formula2


def pairwise_distance(i1, i2, distance, threshold=None):
    """Pairwise distances."""
    for c1, c2 in product(i1, i2):
        score = distance(c1, c2)
        if threshold is None or score < threshold:
            yield (c1, c2), score


def levenshtein(s1, s2):
    """Edit distance function."""
    if len(s1) < len(s2):
        return levenshtein(s2, s1)

    # len(s1) >= len(s2)
    if len(s2) == 0:
        return len(s1)

    previous_row = list(range(len(s2) + 1))
    for i, c1 in enumerate(s1.lower()):
        current_row = [i + 1]
        for j, c2 in enumerate(s2.lower()):
            # We use j+1 instead of j since previous_row and current_row
            # are one character longer than s2
            insertions = previous_row[j + 1] + 1
            deletions = current_row[j] + 1
            substitutions = previous_row[j] + (c1 != c2)
            current_row.append(min(insertions, deletions, substitutions))
        previous_row = current_row

    return previous_row[-1]


def jaccard(s1, s2):
    """Jaccard similarity function."""
    return len(set(s1) & set(s2)) / len(set(s1) | set(s2))



# This file is part of PSAMM.
#
# PSAMM is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PSAMM is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PSAMM.  If not, see <http://www.gnu.org/licenses/>.
#
# Copyright 2017  Jon Lund Steffensen <jon_steffensen@uri.edu>
# Copyright 2015-2020  Keith Dufault-Thompson <keitht547@my.uri.edu>

import unittest

from psamm.metabolicmodel import MetabolicModel
from psamm.database import DictDatabase
from psamm import moma
from psamm.datasource.reaction import parse_reaction
from psamm.lpsolver import generic


class TestLinearMOMA(unittest.TestCase):
    def setUp(self):
        self.database = DictDatabase()
        self.database.set_reaction('rxn_1', parse_reaction('=> (2) |A|'))
        self.database.set_reaction('rxn_2', parse_reaction('|A| <=> |B|'))
        self.database.set_reaction('rxn_3', parse_reaction('|A| => |D|'))
        self.database.set_reaction('rxn_4', parse_reaction('|A| => |C|'))
        self.database.set_reaction('rxn_5', parse_reaction('|C| => |D|'))
        self.database.set_reaction('rxn_6', parse_reaction('|D| =>'))
        self.model = MetabolicModel.load_model(
            self.database, self.database.reactions)

        try:
            self.solver = generic.Solver()
        except generic.RequirementsError:
            self.skipTest('Unable to find an LP solver for tests')

    def test_moma_fba(self):
        p = moma.MOMAProblem(self.model, self.solver)
        fluxes = p.get_fba_flux('rxn_6')
        self.assertAlmostEqual(fluxes['rxn_1'], 500)
        self.assertAlmostEqual(fluxes['rxn_2'], 0)
        self.assertAlmostEqual(fluxes['rxn_6'], 1000)

    def test_moma_minimal_fba(self):
        p = moma.MOMAProblem(self.model, self.solver)
        fluxes = p.get_minimal_fba_flux('rxn_6')
        self.assertAlmostEqual(fluxes['rxn_1'], 500)
        self.assertAlmostEqual(fluxes['rxn_2'], 0)
        self.assertAlmostEqual(fluxes['rxn_3'], 1000)
        self.assertAlmostEqual(fluxes['rxn_6'], 1000)

    def test_linear_moma(self):
        p = moma.MOMAProblem(self.model, self.solver)
        with p.constraints(p.get_flux_var('rxn_3') == 0):
            p.lin_moma({
                'rxn_3': 1000,
                'rxn_4': 0,
                'rxn_5': 0,
            })

        # The closest solution when these are constrained is for
        # rxn_6 to take on a flux of zero.
        self.assertAlmostEqual(p.get_flux('rxn_6'), 0)

    def test_linear_moma2(self):
        p = moma.MOMAProblem(self.model, self.solver)
        with p.constraints(p.get_flux_var('rxn_3') == 0):
            p.lin_moma2('rxn_6', 1000)

        self.assertAlmostEqual(p.get_flux('rxn_1'), 500)
        self.assertAlmostEqual(p.get_flux('rxn_2'), 0)
        self.assertAlmostEqual(p.get_flux('rxn_3'), 0)
        self.assertAlmostEqual(p.get_flux('rxn_4'), 1000)
        self.assertAlmostEqual(p.get_flux('rxn_5'), 1000)
        self.assertAlmostEqual(p.get_flux('rxn_6'), 1000)


class TestQuadraticMOMA(unittest.TestCase):
    def setUp(self):
        self.database = DictDatabase()
        self.database.set_reaction('rxn_1', parse_reaction('=> (2) |A|'))
        self.database.set_reaction('rxn_2', parse_reaction('|A| <=> |B|'))
        self.database.set_reaction('rxn_3', parse_reaction('|A| => |D|'))
        self.database.set_reaction('rxn_4', parse_reaction('|A| => |C|'))
        self.database.set_reaction('rxn_5', parse_reaction('|C| => |D|'))
        self.database.set_reaction('rxn_6', parse_reaction('|D| =>'))
        self.model = MetabolicModel.load_model(
            self.database, self.database.reactions)

        try:
            self.solver = generic.Solver(quadratic=True)
        except generic.RequirementsError:
            self.skipTest('Unable to find an MIQP solver for tests')

    def test_quadratic_moma(self):
        p = moma.MOMAProblem(self.model, self.solver)
        with p.constraints(p.get_flux_var('rxn_3') == 0):
            p.moma({
                'rxn_3': 1000,
                'rxn_4': 0,
                'rxn_5': 0,
            })

        # The closest solution when these are constrained is for
        # rxn_6 to take on a flux of zero.
        self.assertAlmostEqual(p.get_flux('rxn_6'), 0)

    def test_quadratic_moma2(self):
        p = moma.MOMAProblem(self.model, self.solver)
        with p.constraints(p.get_flux_var('rxn_3') == 0):
            p.moma2('rxn_6', 1000)

        self.assertAlmostEqual(p.get_flux('rxn_1'), 500)
        self.assertAlmostEqual(p.get_flux('rxn_2'), 0)
        self.assertAlmostEqual(p.get_flux('rxn_3'), 0)
        self.assertAlmostEqual(p.get_flux('rxn_4'), 1000)
        self.assertAlmostEqual(p.get_flux('rxn_5'), 1000)
        self.assertAlmostEqual(p.get_flux('rxn_6'), 1000)


if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python
# This file is part of PSAMM.
#
# PSAMM is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PSAMM is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PSAMM.  If not, see <http://www.gnu.org/licenses/>.
#
# Copyright 2015-2017  Jon Lund Steffensen <jon_steffensen@uri.edu>
# Copyright 2015-2020  Keith Dufault-Thompson <keitht547@my.uri.edu>

import math
import unittest

from psamm.lpsolver import lp

# Beware: Expressions cannot be tested for equality using normal
# assertEqual because the equality operator is overloaded!


class DummyRangedPropertyContainer(object):
    def __init__(self):
        self._property_rw = 5

    @lp.ranged_property(min=-5, max=10)
    def property_ro(self):
        return 2.0

    @lp.ranged_property(min=-50, max=17)
    def property_rw(self):
        return self._property_rw

    @property_rw.setter
    def property_rw(self, value):
        self._property_rw = value

    @property_rw.deleter
    def property_rw(self):
        self._property_rw = 0


class TestRangedProperty(unittest.TestCase):
    def setUp(self):
        self.c = DummyRangedPropertyContainer()

    def test_read_property_ro(self):
        self.assertEqual(self.c.property_ro.value, 2.0)

    def test_write_property_ro(self):
        with self.assertRaises(AttributeError):
            self.c.property_ro.value = 3.0

    def test_write_property_base(self):
        with self.assertRaises(AttributeError):
            self.c.property_ro = 3.0

    def test_delete_property_base(self):
        with self.assertRaises(AttributeError):
            del self.c.property_ro

    def test_delete_property_ro(self):
        with self.assertRaises(AttributeError):
            del self.c.property_ro.value

    def test_read_property_ro_bounds(self):
        self.assertEqual(self.c.property_ro.min, -5)
        self.assertEqual(self.c.property_ro.max, 10)

    def test_write_property_ro_bounds(self):
        with self.assertRaises(AttributeError):
            self.c.property_ro.min = -50
        with self.assertRaises(AttributeError):
            self.c.property_ro.max = 100

    def test_read_property_rw(self):
        self.assertEqual(self.c.property_rw.value, 5)

    def test_write_property_rw(self):
        self.c.property_rw.value = 10
        self.assertEqual(self.c.property_rw.value, 10)

    def test_delete_property_rw(self):
        del self.c.property_rw.value
        self.assertEqual(self.c.property_rw.value, 0)


class TestExpression(unittest.TestCase):
    def test_create_expression(self):
        e = lp.Expression()
        self.assertEqual(e.offset, 0)
        self.assertEqual(dict(e.values()), {})

    def test_create_expression_with_offset(self):
        e = lp.Expression(offset=42)
        self.assertEqual(e.offset, 42)
        self.assertEqual(dict(e.values()), {})

    def test_create_expression_with_variables(self):
        e = lp.Expression(variables={'x1': -5, 'x2': 100})
        self.assertEqual(e.offset, 0)
        self.assertEqual(dict(e.values()), {'x1': -5, 'x2': 100})

    def test_expression_variables(self):
        e = lp.Expression(variables={'x1': -1, 'x2': 1}, offset=100)
        self.assertEqual(set(e.variables()), {'x1', 'x2'})

    def test_add_expression_and_number(self):
        e = lp.Expression({'x1': -5}, 100)
        e1 = e + 45
        self.assertEqual(e1.offset, 145)
        self.assertEqual(dict(e1.values()), {'x1': -5})

    def test_add_expression_and_expression(self):
        e1 = lp.Expression({'x1': -5, 'x2': 42}, -1000)
        e2 = lp.Expression({'x2': -42, 'x3': -50}, 2)
        e3 = e1 + e2
        self.assertEqual(e3.offset, -998)
        self.assertEqual(dict(e3.values()), {'x1': -5, 'x2': 0, 'x3': -50})

    def test_add_expression_and_infinity(self):
        e = lp.Expression({'x1': -5}, 100)
        e1 = e + float('inf')
        self.assertEqual(e1.offset, float('inf'))

        e2 = e - float('inf')
        self.assertEqual(e2.offset, -float('inf'))

    def test_add_expression_and_infinity_expression(self):
        e = lp.Expression({'x1': -5}, float('inf'))
        e1 = e + lp.Expression({'x2': 4}, 100)
        self.assertEqual(e1.offset, float('inf'))

    def test_add_expression_and_number_in_place(self):
        e = lp.Expression({'x1': -5}, 100)
        e += 45
        self.assertEqual(e.offset, 145)
        self.assertEqual(dict(e.values()), {'x1': -5})

    def test_add_expression_and_expression_in_place(self):
        e1 = lp.Expression({'x1': -5, 'x2': 42}, -1000)
        e2 = lp.Expression({'x2': -42, 'x3': -50}, 2)
        e1 += e2

        self.assertEqual(e1.offset, -998)
        self.assertEqual(dict(e1.values()), {'x1': -5, 'x2': 0, 'x3': -50})

        self.assertEqual(e2.offset, 2)
        self.assertEqual(dict(e2.values()), {'x2': -42, 'x3': -50})

    def test_add_expression_and_infinity_in_place(self):
        e = lp.Expression({'x1': -5}, 100)
        e += float('inf')
        self.assertEqual(e.offset, float('inf'))

    def test_subtract_expressions_in_place(self):
        e = lp.Expression({'x1': 40}, 22)
        e -= lp.Expression({'x1': 5, 'x2': -20}, -2)
        self.assertEqual(e.offset, 24)
        self.assertEqual(dict(e.values()), {'x1': 35, 'x2': 20})

    def test_multiply_expression_and_number(self):
        e = lp.Expression({'x1': -5}, 100)
        e1 = 2 * e
        self.assertEqual(e1.offset, 200)
        self.assertEqual(dict(e1.values()), {'x1': -10})

    def test_multiply_expression_and_infinity(self):
        e = lp.Expression({'x1': -5})
        e1 = float('-inf') * e
        self.assertTrue(math.isnan(e1.offset))

    def test_multiply_expression_and_number_in_place(self):
        e = lp.Expression({'x1': -5}, 100)
        e *= 2
        self.assertEqual(e.offset, 200)
        self.assertEqual(dict(e.values()), {'x1': -10})

    def test_multiply_expression_and_infinity_in_place(self):
        e = lp.Expression({'x1': -5})
        e *= float('inf')
        self.assertTrue(math.isnan(e.offset))

    def test_multiply_expression_and_simple_expression(self):
        e1 = lp.Expression({'x1': -5, 'x2': 3}, offset=10)
        e2 = lp.Expression({'x1': 1})
        e = e1 * e2
        self.assertEqual(e.offset, 0)
        self.assertEqual(dict(e.values()), {
            lp.Product(['x1', 'x1']): -5,
            lp.Product(['x1', 'x2']): 3,
            'x1': 10
        })

    def test_multiply_expression_and_expression(self):
        e1 = lp.Expression({'x1': 1, 'x2': 2, 'x3': 3}, offset=4)
        e2 = lp.Expression({'x2': 4, 'x4': 5}, offset=-5)
        e = e1 * e2
        self.assertEqual(e.offset, -20)
        self.assertEqual(dict(e.values()), {
            lp.Product(['x1', 'x2']): 4,
            lp.Product(['x2', 'x2']): 8,
            lp.Product(['x2', 'x3']): 12,
            lp.Product(['x1', 'x4']): 5,
            lp.Product(['x2', 'x4']): 10,
            lp.Product(['x3', 'x4']): 15,
            'x1': -5,
            'x2': 6,
            'x3': -15,
            'x4': 20
        })

    def test_multiply_expression_and_expression_in_place(self):
        e = lp.Expression({'x1': -5, 'x2': 3}, offset=10)
        e *= lp.Expression({'x1': 1})
        self.assertEqual(e.offset, 0)
        self.assertEqual(dict(e.values()), {
            lp.Product(['x1', 'x1']): -5,
            lp.Product(['x1', 'x2']): 3,
            'x1': 10
        })

    def test_expression_pow_negative(self):
        e = lp.Expression({'x1': -5, 'x2': 3}, offset=10)
        with self.assertRaises(ValueError):
            e1 = e**-2

    def test_expression_pow_neagtive_in_place(self):
        e = lp.Expression({'x1': -5, 'x2': 3}, offset=10)
        with self.assertRaises(ValueError):
            e **= -2

    def test_expression_pow_zero(self):
        e = lp.Expression({'x1': -5, 'x2': 3}, offset=10)
        e1 = e**0
        self.assertEqual(e1.offset, 1)
        self.assertEqual(dict(e1.values()), {})

    def test_expression_pow_zero_in_place(self):
        e = lp.Expression({'x1': -5, 'x2': 3}, offset=10)
        e **= 0
        self.assertEqual(e.offset, 1)
        self.assertEqual(dict(e.values()), {})

    def test_expression_pow_one(self):
        e = lp.Expression({'x1': -5, 'x2': 3}, offset=10)
        e1 = e**1
        self.assertEqual(e1.offset, e.offset)
        self.assertEqual(dict(e1.values()), dict(e.values()))

    def test_expression_pow_one_in_place(self):
        e = lp.Expression({'x1': -5, 'x2': 3}, offset=10)
        e **= 1
        self.assertEqual(e.offset, 10)
        self.assertEqual(dict(e.values()), {'x1': -5, 'x2': 3})

    def test_expression_pow_two(self):
        e = lp.Expression({'x1': -5, 'x2': 3}, offset=10)
        e1 = e**2
        e2 = e*e
        self.assertEqual(e1.offset, e2.offset)
        self.assertEqual(dict(e1.values()), dict(e2.values()))

    def test_expression_pow_two_in_place(self):
        e = lp.Expression({'x1': -5, 'x2': 3}, offset=10)
        e **= 2
        self.assertEqual(e.offset, 100)
        self.assertEqual(dict(e.values()), {
            lp.Product(['x1', 'x1']): 25,
            lp.Product(['x1', 'x2']): -30,
            lp.Product(['x2', 'x2']): 9,
            'x1': -100,
            'x2': 60
        })

    def test_negate_expression(self):
        e = lp.Expression({'x1': 52}, -32)
        e1 = -e
        self.assertEqual(e1.offset, 32)
        self.assertEqual(dict(e1.values()), {'x1': -52})

    def test_expression_to_string(self):
        e = lp.Expression({'x1': -4, 'x2': 100, 'x3': 1}, 42)
        self.assertEqual(str(e), '-4*x1 + 100*x2 + x3 + 42')

    def test_expression_with_product_to_string(self):
        e = lp.Expression({'x1': -4, lp.Product(['x1', 'x2']): 2}, 12)
        self.assertEqual(str(e), '-4*x1 + 2*x1*x2 + 12')

    def test_expression_with_tuple_vars_to_string(self):
        e = lp.Expression({('v', 1): 1}, -1)
        self.assertEqual(str(e), "('v', 1) - 1")

    def test_expression_contains(self):
        e = lp.Expression({'x1': 10, 'x2': -5})
        self.assertIn('x1', e)
        self.assertNotIn('x3', e)

    def test_expression_contains_product(self):
        e = lp.Expression({lp.Product(['x1', 'x1']): -1})
        self.assertIn(lp.Product(['x1', 'x1']), e)
        self.assertNotIn('x1', e)


class TestRelation(unittest.TestCase):
    def assertExpressionEqual(self, e1, e2):
        """Assert that expressions are equal."""
        self.assertEqual(e1.offset, e2.offset)
        self.assertEqual(dict(e1.values()), dict(e2.values()))

    def test_create_relation(self):
        e = lp.Expression({'x1': 4})
        r = lp.Relation(lp.RelationSense.Greater, e)
        self.assertExpressionEqual(r.expression, e)
        self.assertEqual(r.sense, lp.RelationSense.Greater)

    def test_relation_with_offset_to_string(self):
        e = lp.Expression({'x1': 4}, -20)
        r = lp.Relation(lp.RelationSense.Less, e)
        self.assertEqual(str(r), '4*x1 <= 20')

    def test_relation_without_offset_to_string(self):
        e = lp.Expression({'x1': 1})
        r = lp.Relation(lp.RelationSense.Equals, e)
        self.assertEqual(str(r), 'x1 == 0')

    def test_chained_relation(self):
        e = lp.Expression({'x1': 1})
        with self.assertRaises(ValueError):
            self.assertFalse(4 <= e <= 10)


class MockResult(lp.Result):
    def __init__(self, values):
        self._values = dict(values)

    @property
    def success(self):
        return True

    @property
    def status(self):
        return 'Success'

    @property
    def unbounded(self):
        return False

    def _has_variable(self, var):
        return var in self._values

    def _get_value(self, var):
        return self._values[var]


class TestResult(unittest.TestCase):
    def test_result_to_bool(self):
        result = MockResult({'x': 3})
        self.assertTrue(bool(result))

    def test_result_get_simple_value(self):
        result = MockResult({'x': 3, 'y': 4})
        self.assertEqual(result.get_value('x'), 3)
        self.assertEqual(result.get_value('y'), 4)

    def test_result_get_invalid_value(self):
        result = MockResult({'x': 3})
        with self.assertRaises(ValueError):
            y = result.get_value('y')

    def test_result_get_expression_value(self):
        result = MockResult({'x': 3, 'y': 4})
        expr = lp.Expression({'x': 2, 'y': -6}, offset=3)
        self.assertEqual(result.get_value(expr), -15)

    def test_result_get_expression_product_value(self):
        result = MockResult({'x1': 2, 'x2': 3, 'x3': 4})
        expr = lp.Expression({lp.Product(['x1', 'x2']): 3, 'x3': -1}, offset=4)
        self.assertEqual(result.get_value(expr), 18)


if __name__ == '__main__':
    unittest.main()



# This file is part of PSAMM.
#
# PSAMM is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PSAMM is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PSAMM.  If not, see <http://www.gnu.org/licenses/>.
#
# Copyright 2014-2017  Jon Lund Steffensen <jon_steffensen@uri.edu>
# Copyright 2015-2020  Keith Dufault-Thompson <keitht547@my.uri.edu>

from __future__ import unicode_literals

import time
import logging
from itertools import product

from six import text_type

from ..command import (Command, MetabolicMixin, LoopRemovalMixin,
                       SolverCommandMixin, ParallelTaskMixin)
from .. import fluxanalysis, fastcore

logger = logging.getLogger(__name__)


class FluxConsistencyCommand(MetabolicMixin, LoopRemovalMixin,
                             SolverCommandMixin, ParallelTaskMixin, Command):
    """Check that reactions are flux consistent in a model.

    A reaction is flux consistent if there exists any steady-state flux
    solution where the flux of the given reaction is non-zero. The
    bounds on the exchange reactions can be removed when performing the
    consistency check by providing the ``--unrestricted`` option.
    """

    _supported_loop_removal = ['none', 'tfba']

    @classmethod
    def init_parser(cls, parser):
        parser.add_argument(
            '--fastcore', help='Enable use of Fastcore algorithm',
            action='store_true')
        parser.add_argument(
            '--reduce-lp',
            help='Try to reduce the number of LP problems to solve',
            action='store_true')
        parser.add_argument(
            '--epsilon', type=float, help='Flux threshold',
            default=1e-5)
        parser.add_argument(
            '--unrestricted', action='store_true',
            help='Remove limits on exchange reactions before checking')
        super(FluxConsistencyCommand, cls).init_parser(parser)

    def run(self):
        """Run flux consistency check command"""

        # Load compound information
        def compound_name(id):
            if id not in self._model.compounds:
                return id
            return self._model.compounds[id].properties.get('name', id)

        epsilon = self._args.epsilon

        if self._args.unrestricted:
            # Allow all exchange reactions with no flux limits
            for reaction in self._mm.reactions:
                if self._mm.is_exchange(reaction):
                    del self._mm.limits[reaction].bounds

        loop_removal = self._get_loop_removal_option()
        enable_tfba = loop_removal == 'tfba'
        enable_fastcore = self._args.fastcore

        if enable_tfba and enable_fastcore:
            self.argument_error(
                'Using Fastcore with thermodynamic constraints'
                ' is not supported!')
        start_time = time.time()

        if enable_fastcore:
            solver = self._get_solver()
            try:
                inconsistent = set(fastcore.fastcc(
                    self._mm, epsilon, solver=solver))
            except fluxanalysis.FluxBalanceError as e:
                self.report_flux_balance_error(e)
        else:
            if enable_tfba:
                solver = self._get_solver(integer=True)
            else:
                solver = self._get_solver()

            if self._args.reduce_lp:
                logger.info('Running with reduced number of LP problems.')
                try:
                    inconsistent = set(
                        fluxanalysis.consistency_check(
                            self._mm, self._mm.reactions, epsilon,
                            tfba=enable_tfba, solver=solver))
                except fluxanalysis.FluxBalanceError as e:
                    self.report_flux_balance_error(e)
            else:
                logger.info('Using flux bounds to determine consistency.')
                try:
                    inconsistent = set(self._run_fva_fluxcheck(
                        self._mm, solver, enable_tfba, epsilon))
                except FluxCheckFVATaskError:
                    self.report_flux_balance_error()

        logger.info('Solving took {:.2f} seconds'.format(
            time.time() - start_time))

        # Count the number of reactions that are fixed at zero. While these
        # reactions are still inconsistent, they are inconsistent because they
        # have been explicitly disabled.
        disabled_exchange = 0
        disabled_internal = 0

        count_exchange = 0
        total_exchange = 0

        count_internal = 0
        total_internal = 0

        # Print result
        for reaction in sorted(self._mm.reactions):
            disabled = self._mm.limits[reaction].bounds == (0, 0)

            if self._mm.is_exchange(reaction):
                total_exchange += 1
                count_exchange += int(reaction in inconsistent)
                disabled_exchange += int(disabled)
            else:
                total_internal += 1
                count_internal += int(reaction in inconsistent)
                disabled_internal += int(disabled)

            if reaction in inconsistent:
                rx = self._mm.get_reaction(reaction)
                rxt = rx.translated_compounds(compound_name)
                print('{}\t{}'.format(reaction, rxt))

        logger.info('Model has {}/{} inconsistent internal reactions'
                    ' ({} disabled by user)'.format(
                        count_internal, total_internal, disabled_internal))
        logger.info('Model has {}/{} inconsistent exchange reactions'
                    ' ({} disabled by user)'.format(
                        count_exchange, total_exchange, disabled_exchange))

    def _run_fva_fluxcheck(self, model, solver, enable_tfba, epsilon):
        handler_args = model, solver, enable_tfba
        executor = self._create_executor(
            FluxCheckFVATaskHandler, handler_args, cpus_per_worker=2)

        results = {}
        with executor:
            for (reaction_id, direction), value in executor.imap_unordered(
                    product(model.reactions, (1, -1)), 16):

                if reaction_id not in results:
                    results[reaction_id] = value
                    continue

                other_value = results[reaction_id]
                if direction == -1:
                    bounds = value, other_value
                else:
                    bounds = other_value, value

                lower, upper = bounds
                if abs(lower) < epsilon and abs(upper) < epsilon:
                    yield reaction_id

        executor.join()


class FluxCheckFVATaskError(Exception):
    """Error raised from parallel flux check task on failure."""


class FluxCheckFVATaskHandler(object):
    def __init__(self, model, solver, enable_tfba):
        self._problem = fluxanalysis.FluxBalanceProblem(model, solver)
        if enable_tfba:
            self._problem.add_thermodynamic()

    def handle_task(self, reaction_id, direction):
        try:
            return self._problem.flux_bound(reaction_id, direction)
        except fluxanalysis.FluxBalanceError as e:
            # FluxBalanceError is not picklable. Reraise as picklable
            # exception.
            raise FluxCheckFVATaskError(text_type(e))



# This file is part of PSAMM.
#
# PSAMM is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PSAMM is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PSAMM.  If not, see <http://www.gnu.org/licenses/>.
#
# Copyright 2014-2015  Jon Lund Steffensen <jon_steffensen@uri.edu>
# Copyright 2015-2020  Keith Dufault-Thompson <keitht547@my.uri.edu>

"""Representations of boolean expressions and variables.

These classes can be used to represent simple boolean
expressions and do evaluation with substitutions of
particular variables.
"""

from __future__ import unicode_literals

import re

import six
from six import text_type, string_types

from ..util import FrozenOrderedSet


@six.python_2_unicode_compatible
class Variable(object):
    """Represents a variable in a boolean expression"""

    def __init__(self, symbol):
        self._symbol = text_type(symbol)

    @property
    def symbol(self):
        return self._symbol

    def __repr__(self):
        return str('Variable({})').format(repr(self._symbol))

    def __str__(self):
        return self._symbol

    def __eq__(self, other):
        return isinstance(other, Variable) and self._symbol == other._symbol

    def __ne__(self, other):
        return not self == other

    def __hash__(self):
        return hash('Variable') ^ hash(self._symbol)


class _OperatorTerm(object):
    """Composite operator term."""
    def __init__(self, *args):
        terms = list()
        for arg in args:
            if isinstance(arg, self.__class__):
                terms.extend(arg)
            elif isinstance(arg, (bool, Variable, _OperatorTerm)):
                terms.append(arg)
            else:
                raise ValueError('Invalid term: {!r}'.format(arg))
        self._terms = FrozenOrderedSet(terms)

    def __iter__(self):
        return iter(self._terms)

    def __hash__(self):
        return hash(self._terms)

    def __len__(self):
        return len(self._terms)

    def __eq__(self, other):
        return (isinstance(other, self.__class__) and
                self._terms == other._terms)

    def __ne__(self, other):
        return not self == other


class And(_OperatorTerm):
    """Represents an AND term in an expression."""


class Or(_OperatorTerm):
    """Represents an OR term in an expression."""


class SubstitutionError(Exception):
    """Error substituting into expression."""


@six.python_2_unicode_compatible
class Expression(object):
    """Boolean expression representation.

    The expression can be constructed from an expression string of
    variables, operators ("and", "or") and parenthesis groups. For example,

    >>> e = Expression('a and (b or c)')
    """

    def __init__(self, arg, _vars=None):
        if isinstance(arg, (_OperatorTerm, Variable, bool)):
            self._root = arg
        elif isinstance(arg, string_types):
            self._root = _parse_expression(arg)
        else:
            raise TypeError('Unexpected arguments to Expression: {}'.format(
                repr(arg)))

        # If present use _vars to create the set of variables directly. This
        # saves a loop over the tree nodes when the variables are already
        # known.
        if _vars is None:
            variables = []
            if isinstance(self._root, (_OperatorTerm, Variable)):
                stack = [self._root]
                while len(stack) > 0:
                    term = stack.pop()
                    if isinstance(term, Variable):
                        variables.append(term)
                    elif isinstance(term, _OperatorTerm):
                        stack.extend(reversed(list(term)))
                    else:
                        raise ValueError(
                            'Invalid node in expression tree: {!r}'.format(
                                term))

            self._variables = FrozenOrderedSet(variables)
        else:
            self._variables = FrozenOrderedSet(_vars)

    @property
    def root(self):
        """Return root term, variable or boolean of the expression."""
        return self._root

    @property
    def variables(self):
        """Immutable set of variables in the expression."""
        return self._variables

    def has_value(self):
        """Return True if the expression has no variables."""
        return isinstance(self._root, bool)

    @property
    def value(self):
        """The value of the expression if fully evaluated."""
        if not self.has_value():
            raise ValueError('Expression is not fully evaluated')
        return self._root

    def substitute(self, mapping):
        """Substitute variables using mapping function."""
        next_terms = iter([self._root])
        output_stack = []
        current_type = None
        terms = []
        variables = []
        term = None

        while True:
            try:
                term = next(next_terms)
            except StopIteration:
                term = None

            if term is None:
                if current_type is None:
                    term = terms[0]
                    break
                else:
                    if len(terms) == 0:
                        if current_type == And:
                            term = True
                        elif current_type == Or:
                            term = False
                    elif len(terms) == 1:
                        term = terms[0]
                    else:
                        term = current_type(*terms)
                current_type, next_terms, terms = output_stack.pop()
            else:
                if isinstance(term, _OperatorTerm):
                    output_stack.append((current_type, next_terms, terms))
                    current_type = term.__class__
                    terms = []
                    next_terms = iter(term)
                    continue

            # Substitute variable
            if isinstance(term, Variable):
                term = mapping(term)
                if not isinstance(term, (_OperatorTerm, Variable, bool)):
                    raise SubstitutionError(
                        'Expected Variable or bool from substitution,'
                        ' got: {!r}'.format(term))

            # Check again after substitution
            if isinstance(term, Variable):
                variables.append(term)

            # Short circuit with booleans
            while isinstance(term, bool):
                if current_type == And:
                    if not term:
                        current_type, next_terms, terms = output_stack.pop()
                        continue
                    else:
                        break
                elif current_type == Or:
                    if term:
                        current_type, next_terms, terms = output_stack.pop()
                        continue
                    else:
                        break
                else:
                    terms.append(term)
                    break
            else:
                terms.append(term)

        return self.__class__(term, _vars=variables)

    def __repr__(self):
        if isinstance(self._root, bool):
            arg = self._root
        else:
            arg = text_type(self)
        return str('{}({!r})').format(self.__class__.__name__, arg)

    def __str__(self):
        next_terms = iter([self._root])
        output_stack = []
        current_type = None
        terms = []
        term = None

        while True:
            try:
                term = next(next_terms)
            except StopIteration:
                term = None

            if term is None:
                if current_type is None:
                    term = terms[0]
                    break
                elif current_type == And:
                    term = ' and '.join(t for t in terms)
                elif current_type == Or:
                    term = ' or '.join(t for t in terms)
                current_type, next_terms, terms = output_stack.pop()

                # Break on None here to avoid wrapping the outermost term in
                # parentheses.
                if current_type is None:
                    break

                terms.append('(' + term + ')')
            else:
                if isinstance(term, _OperatorTerm):
                    output_stack.append((current_type, next_terms, terms))
                    current_type = term.__class__
                    terms = []
                    next_terms = iter(term)
                else:
                    terms.append(text_type(term))

        return term

    def __eq__(self, other):
        if isinstance(other, self.__class__):
            return self._root == other._root
        return NotImplemented

    def __ne__(self, other):
        if isinstance(other, self.__class__):
            return not self == other
        return NotImplemented


class ParseError(Exception):
    """Signals error parsing boolean expression."""

    def __init__(self, *args, **kwargs):
        self._span = kwargs.pop('span', None)
        super(ParseError, self).__init__(*args, **kwargs)

    @property
    def indicator(self):
        if self._span is None:
            return None
        pre = ' ' * self._span[0]
        ind = '^' * max(1, self._span[1] - self._span[0])
        return pre + ind


def _parse_expression(s):
    """Parse boolean expression containing and/or operators"""

    # Converters for opeartor clauses
    operators = {
        'and': And,
        'or': Or,
        None: lambda *args: args[0]
    }

    # Pairing of end group symbols with start group symbols
    group_pairs = {
        ')': '(',
        ']': '['
    }

    scanner = re.compile(r'''
        (\s+) |              # space
        (\(|\[) |            # group_start
        (\)|\]) |            # group_end
        ((?:or|and)\b) |     # operator
        ([^\s\(\)\[\]]+) |   # variable
        (\Z) |               # end
        (.)                  # error
    ''', re.DOTALL | re.VERBOSE | re.UNICODE | re.IGNORECASE)

    # Parsed using two states and a stack of open clauses
    # At state 0 (not expect_operator): Expect variable, or parenthesis group
    #  start.
    # At state 1 (expect_operator): Expect operator, parenthesis group end, or
    #  end.
    expect_operator = False
    clause_stack = []
    current_clause = []
    clause_operator = None
    clause_symbol = None

    def close():
        prev_op, prev_symbol, prev_clause = clause_stack.pop()
        prev_clause.append(operators[clause_operator](*current_clause))
        return prev_op, prev_symbol, prev_clause

    for match in re.finditer(scanner, s):
        (space, group_start, group_end, operator, variable, end,
            error) = match.groups()

        if error is not None:
            raise ParseError('Invalid token in expression string: {}'.format(
                repr(match.group(0))), span=(match.start(), match.end()))
        elif space is not None:
            continue
        elif expect_operator and operator is not None:
            operator = operator.lower()
            if operator == 'and' and clause_operator != 'and':
                prev_term = current_clause.pop()
                clause_stack.append(
                    (clause_operator, clause_symbol, current_clause))
                current_clause = [prev_term]
            elif operator == 'or' and clause_operator == 'and':
                clause_operator, clause_symbol, current_clause = close()
            clause_operator = operator
            expect_operator = False
        elif expect_operator and group_end is not None:
            if clause_operator == 'and':
                clause_operator, clause_symbol, current_clause = close()
            if len(clause_stack) == 0:
                raise ParseError(
                    'Unbalanced parenthesis group in expression',
                    span=(match.start(), match.end()))
            if group_pairs[group_end] != clause_symbol:
                raise ParseError(
                    'Group started with {} ended with {}'.format(
                        clause_symbol, group_end),
                    span=(match.start(), match.end()))
            clause_operator, clause_symbol, current_clause = close()
        elif expect_operator and end is not None:
            if clause_operator == 'and':
                clause_operator, clause_symbol, current_clause = close()
        elif not expect_operator and variable is not None:
            current_clause.append(Variable(variable))
            expect_operator = True
        elif not expect_operator and group_start is not None:
            clause_stack.append(
                (clause_operator, clause_symbol, current_clause))
            current_clause = []
            clause_operator = None
            clause_symbol = group_start
        else:
            raise ParseError(
                'Invalid token in expression string: {!r}'.format(
                    match.group(0)),
                span=(match.start(), match.end()))

    if len(clause_stack) > 0:
        raise ParseError('Unbalanced parenthesis group in expression')

    expr = operators[clause_operator](*current_clause)
    return expr



# This file is part of PSAMM.
#
# PSAMM is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PSAMM is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PSAMM.  If not, see <http://www.gnu.org/licenses/>.
#
# Copyright 2014-2017  Jon Lund Steffensen <jon_steffensen@uri.edu>
# Copyright 2015-2020  Keith Dufault-Thompson <keitht547@my.uri.edu>
# Copyright 2020-2020  Elysha Sameth <esameth1@my.uri.edu>

"""Module for reading and writing native formats.

These formats are either table-based or YAML-based. Table-based formats
are space-separated and empty lines are ignored. Comments starting with
pound (#). YAML-based formats are structured data following the YAML
specification.
"""

from __future__ import absolute_import, unicode_literals

import os
import logging
import re
import csv
import math
from collections import OrderedDict

import yaml
from six import string_types, text_type, iteritems, itervalues, PY3
from decimal import Decimal

from ..reaction import Reaction, Compound, Direction
from ..expression import boolean
from ..formula import Formula
from ..metabolicmodel import MetabolicModel
from ..database import DictDatabase
from .context import FilePathContext, FileMark
from .entry import (DictCompoundEntry as CompoundEntry,
                    DictReactionEntry as ReactionEntry,
                    DictCompartmentEntry as CompartmentEntry)
from .reaction import ReactionParser, convert_to_unicode
from . import modelseed
from .. import util

# Module-level logging
logger = logging.getLogger(__name__)

_HAS_YAML_LIBRARY = None

_REACTION_PARSER = ReactionParser()


class ParseError(Exception):
    """Exception used to signal errors while parsing"""


def float_constructor(loader, node):
    """Construct Decimal from YAML float encoding."""
    s = loader.construct_scalar(node)
    if s == '.inf':
        return Decimal('Infinity')
    elif s == '-.inf':
        return -Decimal('Infinity')
    elif s == '.nan':
        return Decimal('NaN')
    return Decimal(s)


def yaml_load(stream):
    """Load YAML file using safe loader."""
    # Surprisingly, the CSafeLoader does not seem to be used by default.
    # Check whether the CSafeLoader is available and provide a log message
    # if it is not available.
    global _HAS_YAML_LIBRARY

    if _HAS_YAML_LIBRARY is None:
        _HAS_YAML_LIBRARY = hasattr(yaml, 'CSafeLoader')
        if not _HAS_YAML_LIBRARY:
            logger.warning('libyaml was not found! Please install libyaml to'
                           ' speed up loading the model files.')

    if _HAS_YAML_LIBRARY:
        loader = yaml.CSafeLoader(stream)
    else:
        loader = yaml.SafeLoader(stream)
    loader.add_constructor('tag:yaml.org,2002:float', float_constructor)
    return loader.get_data()


class _OrderedEntrySet(object):
    """Ordered set of entity entries.

    This is somewhere between a Set and a Mapping but deliberately does not
    implement any of those interfaces. Iteration does not provide
    (key, value)-items (since the keys are already in the entry values)
    but instead provides just values. The usual set operations such as union
    and intersection are not provided.
    """
    def __init__(self, mapping={}):
        self._dict = OrderedDict(mapping)

    def __contains__(self, key):
        return key in self._dict

    def get(self, key, default=None):
        try:
            return self.__getitem__(key)
        except KeyError:
            return default

    def __getitem__(self, key):
        return self._dict[key]

    def __iter__(self):
        return itervalues(self._dict)

    def __len__(self):
        return len(self._dict)

    def add_entry(self, entry):
        self._dict[entry.id] = entry

    def discard(self, key):
        try:
            del self._dict[key]
        except KeyError:
            pass

    def clear(self):
        self._dict.clear()

    def update(self, it):
        for entry in it:
            self.add_entry(entry)

    def __repr__(self):
        return str('<_OrderedEntrySet {{{}}}>').format(
            ', '.join(repr(x) for x in iter(self)))


class ModelReader(object):
    """Reader of native YAML-based model format.

    The reader can be created from a model YAML file or directly from a
    dict, string or File-like object. Use :meth:`reader_from_path` to read the
    model from a YAML file or directory and use the constructor to read from
    other sources. Any externally referenced file (with ``include``) will be
    read on demand by the parse methods. To read the model fully into memory,
    use the :meth:`create_model` to create a :class:`NativeModel`.
    """

    # Model files to try to open if a directory was specified
    DEFAULT_MODEL = 'model.yaml', 'model.yml'

    def __init__(self, model_from, context=None):
        """Create a model from the specified content.

        Model can be a string, open file, or dictionary.
        """
        if isinstance(model_from, string_types):
            self._model = yaml_load(model_from)
            self._context = context
        elif isinstance(model_from, dict):
            self._context = context
            self._model = model_from
        elif hasattr(model_from, 'read') and callable(model_from.read):
            self._context = context
            self._model = yaml_load(model_from)
        else:
            raise ValueError("Model is of an invalid types")

    @classmethod
    def reader_from_path(cls, path):
        """Create a model from specified path.

        Path can be a directory containing a ``model.yaml`` or ``model.yml``
        file or it can be a path naming the central model file directly.
        """
        context = FilePathContext(path)
        try:
            with context.open('r') as f:
                return ModelReader(f, context)
        except IOError:
            # Try to open the default file
            for filename in cls.DEFAULT_MODEL:
                try:
                    context = FilePathContext(
                        os.path.join(path, filename))
                    with context.open('r') as f:
                        return ModelReader(f, context)
                except:
                    logger.debug('Failed to load model file',
                                 exc_info=True)

        # No model could be loaded
        raise ParseError('No model file could be found ({})'.format(
            ', '.join(cls.DEFAULT_MODEL)))

    @property
    def name(self):
        """Name specified by the model."""
        return self._model.get('name', None)

    @property
    def biomass_reaction(self):
        """Biomass reaction specified by the model."""
        return self._model.get('biomass', None)

    @property
    def extracellular_compartment(self):
        """Extracellular compartment specified by the model.

        Defaults to 'e'.
        """
        return self._model.get('extracellular', 'e')

    @property
    def default_compartment(self):
        """Default compartment specified by the model.

        The compartment that is implied when not specified. In some contexts
        (e.g. for exchange compounds) the extracellular compartment may be
        implied instead. Defaults to 'c'.
        """
        return self._model.get('default_compartment', 'c')

    @property
    def default_flux_limit(self):
        """Default flux limit specified by the model.

        When flux limits on reactions are not specified, this value will be
        used. Flux limit of [0;x] will be implied for irreversible reactions
        and [-x;x] for reversible reactions, where x is this value.
        Defaults to 1000."""
        return self._model.get('default_flux_limit', 1000)

    def parse_compartments(self):
        """Parse compartment information from model.

        Return tuple of: 1) iterator of
        :class:`psamm.datasource.entry.CompartmentEntry`; 2) Set of pairs
        defining the compartment boundaries of the model.
        """

        compartments = OrderedDict()
        boundaries = set()

        if 'compartments' in self._model:
            boundary_map = {}
            for compartment_def in self._model['compartments']:
                compartment_id = convert_to_unicode(compartment_def.get('id'))
                _check_id(compartment_id, 'Compartment')
                if compartment_id in compartments:
                    raise ParseError('Duplicate compartment ID: {}'.format(
                        compartment_id))
                props = dict(compartment_def)
                adjacent_to = props.pop('adjacent_to', None)
                if adjacent_to is not None:
                    if not isinstance(adjacent_to, list):
                        adjacent_to = [adjacent_to]
                    for other in adjacent_to:
                        boundary_map.setdefault(other, set()).add(
                            compartment_id)

                mark = FileMark(self._context, None, None)
                compartment = CompartmentEntry(props, mark)
                compartments[compartment_id] = compartment

            # Check boundaries from boundary_map
            for source, dest_set in iteritems(boundary_map):
                if source not in compartments:
                    raise ParseError(
                        'Invalid compartment {} referenced'
                        ' by compartment {}'.format(
                            source, ', '.join(dest_set)))
                for dest in dest_set:
                    boundaries.add(tuple(sorted((source, dest))))

        return itervalues(compartments), frozenset(boundaries)

    def parse_reactions(self):
        """Yield tuples of reaction ID and reactions defined in the model"""

        # Parse reactions defined in the main model file
        if 'reactions' in self._model:
            for reaction in parse_reaction_list(
                    self._context, self._model['reactions'],
                    self.default_compartment):
                yield reaction

    def has_model_definition(self):
        """Return True when the list of model reactions is set in the model."""
        return 'model' in self._model

    def parse_model(self):
        """Yield reaction IDs of model reactions"""

        if self.has_model_definition():
            for reaction_id in parse_model_group_list(
                    self._context, self._model['model']):
                yield reaction_id

    def parse_limits(self):
        """Yield tuples of reaction ID, lower, and upper bound flux limits"""

        if 'limits' in self._model:
            if not isinstance(self._model['limits'], list):
                raise ParseError('Expected limits to be a list')

            for limit in parse_limits_list(
                    self._context, self._model['limits']):
                yield limit

    def parse_exchange(self):
        """Yield tuples of exchange compounds.

        Each exchange compound is a tuple of compound, reaction ID, lower and
        upper flux limits.
        """

        if 'media' in self._model:
            if 'exchange' in self._model:
                raise ParseError('Both "media" and "exchange" are specified')
            logger.warning(
                'The "media" key is deprecated! Please use "exchange" instead:'
                ' https://psamm.readthedocs.io/en/stable/file_format.html')
            exchange_list = self._model['media']
        else:
            exchange_list = self._model.get('exchange')

        extracellular = self.extracellular_compartment
        if exchange_list is not None:
            if not isinstance(exchange_list, list):
                raise ParseError('Expected "exchange" to be a list')

            for exchange_compound in parse_exchange_list(
                    self._context, exchange_list, extracellular):
                compound, reaction_id, lower, upper = exchange_compound
                if compound.compartment is None:
                    compound = compound.in_compartment(extracellular)
                yield compound, reaction_id, lower, upper

    parse_medium = parse_exchange
    """Yield tuples of exchange compounds.

    .. deprecated:: 0.28
       Use :meth:`parse_exchange` instead.
    """

    def parse_compounds(self):
        """Yield CompoundEntries for defined compounds"""

        if 'compounds' in self._model:
            for compound in parse_compound_list(
                    self._context, self._model['compounds']):
                yield compound

    @property
    def context(self):
        return self._context

    def create_model(self):
        """Return :class:`NativeModel` fully loaded into memory."""
        properties = {
            'name': self.name,
            'biomass': self.biomass_reaction,
            'extracellular': self.extracellular_compartment,
            'default_compartment': self.default_compartment,
            'default_flux_limit': self.default_flux_limit
        }

        if self.context is not None:
            git_version = util.git_try_describe(self.context.basepath)
            properties['version_string'] = git_version

        model = NativeModel(properties)

        # Load compartments into model
        compartment_iter, boundaries = self.parse_compartments()
        for compartment in compartment_iter:
            model.compartments.add_entry(compartment)
        model.compartment_boundaries.update(boundaries)

        # Load compounds into model
        for compound in self.parse_compounds():
            if compound.id in model.compounds:
                existing_entry = model.compounds[compound.id]
                common_props = set(compound.properties).intersection(
                    existing_entry.properties).difference({'id'})
                if len(common_props) > 0:
                    logger.warning(
                        'Compound entry {} at {} overrides already defined'
                        ' properties: {}'.format(
                            compound.id, compound.filemark, common_props))

                properties = dict(compound.properties)
                properties.update(existing_entry.properties)
                compound = CompoundEntry(
                    properties, filemark=compound.filemark)
            model.compounds.add_entry(compound)

        # Load reactions into model
        for reaction in self.parse_reactions():
            if reaction.id in model.reactions:
                existing_entry = model.reactions[reaction.id]
                common_props = set(reaction.properties).intersection(
                    existing_entry.properties).difference({'id'})
                if len(common_props) > 0:
                    logger.warning(
                        'Reaction entry {} at {} overrides already defined'
                        ' properties: {}'.format(
                            reaction.id, reaction.filemark, common_props))

                properties = dict(reaction.properties)
                properties.update(existing_entry.properties)
                reaction = ReactionEntry(
                    properties, filemark=reaction.filemark)
            model.reactions.add_entry(reaction)

        for exchange_def in self.parse_exchange():
            model.exchange[exchange_def[0]] = exchange_def

        for limit in self.parse_limits():
            model.limits[limit[0]] = limit
            dir = model.reactions.get(limit[0]).properties[
                'equation'].direction
            if str(dir) == 'Direction.Forward' or \
                    str(dir) == 'Direction.Right':
                v_min = 0
                v_max = model.default_flux_limit
            elif str(dir) == 'Direction.Both' or str(dir) == 'Direction.Bidir':
                v_min = -model.default_flux_limit
                v_max = model.default_flux_limit
            else:
                logger.warning('Reaction {} has invalid direction: {}'.format(
                    limit[0], dir))
            if (limit[1] is not None and limit[1] < v_min) or \
                    (limit[2] is not None and limit[2] > v_max):
                specify_vmin = limit[1] if limit[1] is not None else \
                    -model.default_flux_limit
                specify_vmax = limit[2] if limit[2] is not None else \
                    -model.default_flux_limit
                logger.warning(
                    "Flux limits for reaction {} override default flux "
                    "limits set by reaction direction (current flux limits: "
                    "[{}, {}]; default limits: [{}, {}]).".format(
                        limit[0], specify_vmin, specify_vmax, v_min, v_max))

        if self.has_model_definition():
            for model_reaction in self.parse_model():
                model.model[model_reaction] = None
        else:
            for reaction in model.reactions:
                model.model[reaction.id] = None

        return model


class NativeModel(object):
    """Represents model in the native format."""

    def __init__(self, properties={}):
        self._properties = dict(properties)
        self._compartments = _OrderedEntrySet()
        self._compartment_boundaries = set()
        self._compounds = _OrderedEntrySet()
        self._reactions = _OrderedEntrySet()
        self._exchange = OrderedDict()
        self._limits = OrderedDict()
        self._model = OrderedDict()

    @property
    def name(self):
        """Return model name property."""
        return self._properties.get('name')

    @name.setter
    def name(self, value):
        self._properties['name'] = value

    @property
    def version_string(self):
        """Return model version string."""
        return self._properties.get('version_string')

    @version_string.setter
    def version_string(self, value):
        self._properties['version_string'] = value

    @property
    def biomass_reaction(self):
        """Return biomass reaction property."""
        return self._properties.get('biomass')

    @biomass_reaction.setter
    def biomass_reaction(self, value):
        self._properties['biomass'] = value

    @property
    def extracellular_compartment(self):
        """Return extracellular compartment property."""
        return self._properties.get('extracellular')

    @extracellular_compartment.setter
    def extracellular_compartment(self, value):
        self._properties['extracellular'] = value

    @property
    def default_compartment(self):
        """Return default compartment property."""
        return self._properties.get('default_compartment')

    @default_compartment.setter
    def default_compartment(self, value):
        self._properties['default_compartment'] = value

    @property
    def default_flux_limit(self):
        """Return default flux limit property."""
        return self._properties.get('default_flux_limit')

    @default_flux_limit.setter
    def default_flux_limit(self, value):
        self._properties['default_flux_limit'] = value

    @property
    def compartments(self):
        """Return compartments entry set."""
        return self._compartments

    @property
    def compartment_boundaries(self):
        """Return set of compartment boundaries."""
        return self._compartment_boundaries

    @property
    def reactions(self):
        """Return reaction entry set."""
        return self._reactions

    @property
    def compounds(self):
        """Return compound entry set."""
        return self._compounds

    @property
    def exchange(self):
        """Return dict of exchange compounds and properties."""
        return self._exchange

    @property
    def limits(self):
        """Return dict of reaction limits."""
        return self._limits

    @property
    def model(self):
        """Return dict of model reactions."""
        return self._model

    def create_metabolic_model(self):
        """Create a :class:`psamm.metabolicmodel.MetabolicModel`."""

        def _translate_compartments(reaction, compartment):
            """Translate compound with missing compartments.

            These compounds will have the specified compartment in the output.
            """
            left = (((c.in_compartment(compartment), v)
                     if c.compartment is None else (c, v))
                    for c, v in reaction.left)
            right = (((c.in_compartment(compartment), v)
                      if c.compartment is None else (c, v))
                     for c, v in reaction.right)
            return Reaction(reaction.direction, left, right)

        # Create metabolic model
        database = DictDatabase()
        for reaction in self.reactions:
            if reaction.equation is not None:
                equation = _translate_compartments(
                    reaction.equation, self.default_compartment)
                database.set_reaction(reaction.id, equation)

        undefined_compartments = set()
        undefined_compounds = set()
        extracellular_compounds = set()
        extracellular = self.extracellular_compartment
        for reaction in database.reactions:
            for compound, _ in database.get_reaction_values(reaction):
                if compound.name not in self.compounds:
                    undefined_compounds.add(compound.name)
                if compound.compartment == extracellular:
                    extracellular_compounds.add(compound.name)
                if compound.compartment not in self.compartments:
                    undefined_compartments.add(compound.compartment)

        for compartment in sorted(undefined_compartments):
            logger.warning(
                'The compartment {} was not defined in the list'
                ' of compartments'.format(compartment))

        for compound in sorted(undefined_compounds):
            logger.warning(
                'The compound {} was not defined in the list'
                ' of compounds'.format(compound))

        exchange_compounds = set()
        for exchange_compound in self.exchange:
            if exchange_compound.compartment == extracellular:
                exchange_compounds.add(exchange_compound.name)

        for compound in sorted(extracellular_compounds - exchange_compounds):
            logger.warning(
                'The compound {} was in the extracellular compartment'
                ' but not defined in the exchange compounds'.format(compound))
        for compound in sorted(exchange_compounds - extracellular_compounds):
            logger.warning(
                'The compound {} was defined in the exchange compounds but'
                ' is not in the extracellular compartment'.format(compound))

        model_definition = None
        if len(self.model) > 0:
            model_definition = self.model

        for reaction in self.reactions:
            if reaction.equation is None:
                logger.warning(
                    'Reaction {} has no reaction equation'.format(reaction.id))
                del model_definition[reaction.id]

        return MetabolicModel.load_model(
            database, model_definition, itervalues(self.exchange),
            itervalues(self.limits), v_max=self.default_flux_limit)

    def __repr__(self):
        return str('<{} name={!r}>'.format(self.__class__.__name__, self.name))


def _check_id(entity, entity_type):
    """Check whether the ID is valid.

    First check if the ID is missing, and then check if it is a qualified
    string type, finally check if the string is empty. For all checks, it
    would raise a ParseError with the corresponding message.

    Args:
        entity: a string type object to be checked.
        entity_type: a string that shows the type of entities to check, usually
            `Compound` or 'Reaction'.
    """

    if entity is None:
        raise ParseError('{} ID missing'.format(entity_type))
    elif not isinstance(entity, string_types):
        msg = '{} ID must be a string, id was {}.'.format(entity_type, entity)
        if isinstance(entity, bool):
            msg += (' You may have accidentally used an ID value that YAML'
                    ' interprets as a boolean, such as "yes", "no", "on",'
                    ' "off", "true" or "false". To use this ID, you have to'
                    ' quote it with single or double quotes')
        raise ParseError(msg)
    elif len(entity) == 0:
        raise ParseError('{} ID must not be empty'.format(entity_type))


def parse_compound(compound_def, context=None):
    """Parse a structured compound definition as obtained from a YAML file

    Returns a CompoundEntry."""
    compound_id = convert_to_unicode(compound_def.get('id'))
    _check_id(compound_id, 'Compound')

    compound_props = {}
    for key, value in compound_def.items():
        if isinstance(value, str):
            compound_props[key] = convert_to_unicode(value)
        else:
            compound_props[key] = value

    mark = FileMark(context, None, None)
    return CompoundEntry(compound_props, mark)


def parse_compound_list(path, compounds):
    """Parse a structured list of compounds as obtained from a YAML file

    Yields CompoundEntries. Path can be given as a string or a context.
    """
    context = FilePathContext(path)

    for compound_def in compounds:
        compound_dict = dict(compound_def)
        if 'include' in compound_dict:
            file_format = compound_dict.get('format')
            include_context = context.resolve(compound_dict['include'])
            for compound in parse_compound_file(include_context, file_format):
                yield compound
        else:
            yield parse_compound(compound_dict, context)


def parse_compound_table_file(path, f):
    """Parse a tab-separated file containing compound IDs and properties

    The compound properties are parsed according to the header which specifies
    which property is contained in each column.
    """

    context = FilePathContext(path)

    for i, row in enumerate(csv.DictReader(f, delimiter=str('\t'))):
        if text_type('id') not in row or \
                text_type(convert_to_unicode(row['id'].strip())) == '':
            raise ParseError('Expected `id` column in table')

        props = {key: text_type(convert_to_unicode(value))
                 for key, value in iteritems(row)
                 if text_type(convert_to_unicode(value)) != ''}
        if 'charge' in props:
            props['charge'] = int(props['charge'])

        mark = FileMark(context, i + 2, None)
        yield CompoundEntry(props, mark)


def parse_compound_yaml_file(path, f):
    """Parse a file as a YAML-format list of compounds

    Path can be given as a string or a context.
    """
    return parse_compound_list(path, yaml_load(f))


def resolve_format(format, path):
    """Looks at a file's extension and format (if any) and returns format.
    """
    if format is None:
        if (re.match(r'.+\.(yml|yaml)$', path)):
            return 'yaml'
        elif (re.match(r'.+\.tsv$', path)):
            return 'tsv'
    else:
        return format.lower()


def parse_compound_file(path, format):
    """Open and parse reaction file based on file extension or given format

    Path can be given as a string or a context.
    """

    context = FilePathContext(path)

    # YAML files do not need to explicitly specify format
    format = resolve_format(format, context.filepath)
    if format == 'yaml':
        logger.debug('Parsing compound file {} as YAML'.format(
            context.filepath))
        with context.open('r') as f:
            for compound in parse_compound_yaml_file(context, f):
                yield compound
    elif format == 'modelseed':
        logger.debug('Parsing compound file {} as ModelSEED TSV'.format(
            context.filepath))
        with context.open('r') as f:
            for compound in modelseed.parse_compound_file(f, context):
                yield compound
    elif format == 'tsv':
        logger.debug('Parsing compound file {} as TSV'.format(
            context.filepath))
        with context.open('r') as f:
            for compound in parse_compound_table_file(context, f):
                yield compound
    else:
        raise ParseError('Unable to detect format of compound file {}'.format(
            context.filepath))


def parse_reaction_equation_string(equation, default_compartment):
    """Parse a string representation of a reaction equation.

    Converts undefined compartments to the default compartment.
    """
    def _translate_compartments(reaction, compartment):
        """Translate compound with missing compartments.

        These compounds will have the specified compartment in the output.
        """
        left = (((c.in_compartment(compartment), v)
                 if c.compartment is None else (c, v))
                for c, v in reaction.left)
        right = (((c.in_compartment(compartment), v)
                  if c.compartment is None else (c, v))
                 for c, v in reaction.right)
        return Reaction(reaction.direction, left, right)
    eq = _REACTION_PARSER.parse(equation).normalized()
    return _translate_compartments(eq, default_compartment)


def parse_reaction_equation(equation_def, default_compartment):
    """Parse a structured reaction equation as obtained from a YAML file

    Returns a Reaction.
    """
    def parse_compound_list(compound_list, compartment):
        """Parse a list of reactants or metabolites"""
        for compound_def in compound_list:
            compound_id = convert_to_unicode(compound_def.get('id'))
            _check_id(compound_id, 'Compound')

            value = compound_def.get('value')
            if value is None:
                raise ParseError('Missing value for compound {}'.format(
                    compound_id))

            compound_compartment = compound_def.get('compartment')
            if compound_compartment is None:
                compound_compartment = compartment

            compound = Compound(compound_id, compartment=compound_compartment)
            yield compound, value

    if isinstance(equation_def, string_types):
        return parse_reaction_equation_string(
            equation_def, default_compartment)
    else:
        compartment = equation_def.get('compartment', default_compartment)
        reversible = bool(equation_def.get('reversible', True))
        left = equation_def.get('left', [])
        right = equation_def.get('right', [])
        if len(left) == 0 and len(right) == 0:
            raise ParseError('Reaction values are missing')

        return Reaction(Direction.Both if reversible else Direction.Forward,
                        parse_compound_list(left, compartment),
                        parse_compound_list(right, compartment))


def parse_reaction(reaction_def, default_compartment, context=None):
    """Parse a structured reaction definition as obtained from a YAML file

    Returns a ReactionEntry.
    """
    reaction_id = convert_to_unicode(reaction_def.get('id'))
    _check_id(reaction_id, 'Reaction')

    reaction_props = {}
    for key, value in reaction_def.items():
        if isinstance(value, str):
            reaction_props[key] = convert_to_unicode(value)
        else:
            reaction_props[key] = value

    if 'genes' in reaction_def.keys():
        if isinstance(reaction_def['genes'], list):
            reaction_props['genes'] = [convert_to_unicode(gene)
                                       for gene in reaction_def['genes']]

    # Parse reaction equation
    if 'equation' in reaction_def.keys():
        reaction_props['equation'] = parse_reaction_equation(
            reaction_def['equation'], default_compartment)

    mark = FileMark(context, None, None)
    return ReactionEntry(reaction_props, mark)


def parse_reaction_list(path, reactions, default_compartment=None):
    """Parse a structured list of reactions as obtained from a YAML file

    Yields tuples of reaction ID and reaction object. Path can be given as a
    string or a context.
    """
    context = FilePathContext(path)

    for reaction_def in reactions:
        reaction_dict = dict(reaction_def)
        if 'include' in reaction_dict:
            include_context = context.resolve(reaction_dict['include'])
            for reaction in parse_reaction_file(
                    include_context, default_compartment):
                yield reaction
        else:
            yield parse_reaction(reaction_dict, default_compartment, context)


def parse_reaction_yaml_file(path, f, default_compartment):
    """Parse a file as a YAML-format list of reactions

    Path can be given as a string or a context.
    """

    return parse_reaction_list(path, yaml_load(f), default_compartment)


def parse_reaction_table_file(path, f, default_compartment):
    """Parse a tab-separated file containing reaction IDs and properties

    The reaction properties are parsed according to the header which specifies
    which property is contained in each column.
    """

    context = FilePathContext(path)

    for lineno, row in enumerate(csv.DictReader(f, delimiter=str('\t'))):
        if text_type('id') not in row or \
                text_type(convert_to_unicode(row['id'].strip())) == '':
            raise ParseError('Expected `id` column in table')

        props = {key: convert_to_unicode(value)
                 for key, value in iteritems(row)
                 if convert_to_unicode(value) != ''}

        if 'equation' in props:
            props['equation'] = parse_reaction_equation_string(
                props['equation'], default_compartment)

        mark = FileMark(context, lineno + 2, 0)
        yield ReactionEntry(props, mark)


def parse_reaction_file(path, default_compartment=None):
    """Open and parse reaction file based on file extension

    Path can be given as a string or a context.
    """

    context = FilePathContext(path)

    format = resolve_format(None, context.filepath)
    if format == 'tsv':
        logger.debug('Parsing reaction file {} as TSV'.format(
            context.filepath))
        with context.open('r') as f:
            for reaction in parse_reaction_table_file(
                    context, f, default_compartment):
                yield reaction
    elif format == 'yaml':
        logger.debug('Parsing reaction file {} as YAML'.format(
            context.filepath))
        with context.open('r') as f:
            for reaction in parse_reaction_yaml_file(
                    context, f, default_compartment):
                yield reaction
    else:
        raise ParseError('Unable to detect format of reaction file {}'.format(
            context.filepath))


def get_limits(compound_def):
    if ('fixed' in compound_def and
            ('lower' not in compound_def and 'upper'not in compound_def)):
        fixed = compound_def['fixed']
        lower = fixed
        upper = fixed
    elif ('fixed' in compound_def and
            ('lower' in compound_def or 'upper' in compound_def)):
        raise ParseError('Cannot use fixed and a lower or upper bound')
    else:
        lower = compound_def.get('lower', None)
        upper = compound_def.get('upper', None)
    return lower, upper


def parse_exchange(exchange_def, default_compartment):
    """Parse a structured exchange definition as obtained from a YAML file.

    Returns in iterator of compound, reaction, lower and upper bounds.
    """

    default_compartment = exchange_def.get('compartment', default_compartment)

    for compound_def in exchange_def.get('compounds', []):
        compartment = compound_def.get('compartment', default_compartment)
        compound = Compound(convert_to_unicode(compound_def['id']),
                            compartment=compartment)
        reaction = compound_def.get('reaction')
        if reaction:
            reaction = convert_to_unicode(reaction)
        lower, upper = get_limits(compound_def)
        yield compound, reaction, lower, upper


parse_medium = parse_exchange
"""Parse a structured exchange definition as obtained from a YAML file.

.. deprecated:: 0.28
   Use :func:`parse_exchange` instead.
"""


def parse_exchange_list(path, exchange, default_compartment):
    """Parse a structured exchange list as obtained from a YAML file.

    Yields tuples of compound, reaction ID, lower and upper flux bounds. Path
    can be given as a string or a context.
    """

    context = FilePathContext(path)

    for exchange_def in exchange:
        if 'include' in exchange_def:
            include_context = context.resolve(exchange_def['include'])
            for exchange_compound in parse_exchange_file(
                    include_context, default_compartment):
                yield exchange_compound
        else:
            for exchange_compound in parse_exchange(
                    exchange_def, default_compartment):
                yield exchange_compound


parse_medium_list = parse_exchange_list
"""Parse a structured exchange list as obtained from a YAML file.

.. deprecated:: 0.28
   Use :func:`parse_exchange_list` instead.
"""


def parse_exchange_yaml_file(path, f, default_compartment):
    """Parse a file as a YAML-format exchange definition.

    Path can be given as a string or a context.
    """

    return parse_exchange(yaml_load(f), default_compartment)


parse_medium_yaml_file = parse_exchange_yaml_file
"""Parse a file as a YAML-format exchange definition.

.. deprecated:: 0.28
   Use :func:`parse_exchange_yaml_file` instead.
"""


def parse_exchange_table_file(f):
    """Parse a space-separated file containing exchange compound flux limits.

    The first two columns contain compound IDs and compartment while the
    third column contains the lower flux limits. The fourth column is
    optional and contains the upper flux limit.
    """

    for line in f:
        line, _, comment = convert_to_unicode(line).partition('#')
        line = line.strip()
        if line == '':
            continue

        # A line can specify lower limit only (useful for
        # medium compounds), or both lower and upper limit.
        fields = line.split(None)
        if len(fields) < 2 or len(fields) > 4:
            raise ParseError('Malformed compound limit: {}'.format(fields))

        # Extend to four values and unpack
        fields.extend(['-']*(4-len(fields)))
        compound_id, compartment, lower, upper = fields

        compound = Compound(convert_to_unicode(compound_id), compartment)
        lower = float(lower) if lower != '-' else None
        upper = float(upper) if upper != '-' else None

        yield compound, None, lower, upper


parse_medium_table_file = parse_exchange_table_file
"""Parse a space-separated file containing exchange compound flux limits.

.. deprecated:: 0.28
   Use :func:`parse_exchange_table_file` instead.
"""


def parse_exchange_file(path, default_compartment):
    """Parse a file as a list of exchange compounds with flux limits.

    The file format is detected and the file is parsed accordingly. Path can
    be given as a string or a context.
    """

    context = FilePathContext(path)

    format = resolve_format(None, context.filepath)
    if format == 'tsv':
        logger.debug('Parsing exchange file {} as TSV'.format(
            context.filepath))
        with context.open('r') as f:
            for entry in parse_exchange_table_file(f):
                yield entry
    elif format == 'yaml':
        logger.debug('Parsing exchange file {} as YAML'.format(
            context.filepath))
        with context.open('r') as f:
            for entry in parse_exchange_yaml_file(
                    context, f, default_compartment):
                yield entry
    else:
        raise ParseError('Unable to detect format of exchange file {}'.format(
            context.filepath))


parse_medium_file = parse_exchange_file
"""Parse a file as a list of exchange compounds with flux limits.

.. deprecated:: 0.28
   Use :func:`parse_exchange_file` instead.
"""


def parse_limit(limit_def):
    """Parse a structured flux limit definition as obtained from a YAML file

    Returns a tuple of reaction, lower and upper bound.
    """

    lower, upper = get_limits(limit_def)
    reaction = convert_to_unicode(limit_def.get('reaction'))

    return reaction, lower, upper


def parse_limits_list(path, limits):
    """Parse a structured list of flux limits as obtained from a YAML file

    Yields tuples of reaction ID, lower and upper flux bounds. Path can be
    given as a string or a context.
    """

    context = FilePathContext(path)

    for limit_def in limits:
        if 'include' in limit_def:
            include_context = context.resolve(limit_def['include'])
            for limit in parse_limits_file(include_context):
                yield limit
        else:
            yield parse_limit(limit_def)


def parse_limits_table_file(f):
    """Parse a space-separated file containing reaction flux limits

    The first column contains reaction IDs while the second column contains
    the lower flux limits. The third column is optional and contains the
    upper flux limit.
    """

    for line in f:
        line, _, comment = convert_to_unicode(line).partition('#')
        line = line.strip()
        if line == '':
            continue

        # A line can specify lower limit only (useful for
        # exchange reactions), or both lower and upper limit.
        fields = line.split(None)
        if len(fields) < 1 or len(fields) > 3:
            raise ParseError('Malformed reaction limit: {}'.format(fields))

        # Extend to three values and unpack
        fields.extend(['-']*(3-len(fields)))
        reaction_id, lower, upper = fields

        lower = float(lower) if lower != '-' else None
        upper = float(upper) if upper != '-' else None

        yield convert_to_unicode(reaction_id), lower, upper


def parse_limits_yaml_file(path, f):
    """Parse a file as a YAML-format flux limits definition

    Path can be given as a string or a context.
    """

    return parse_limits_list(path, yaml_load(f))


def parse_limits_file(path):
    """Parse a file as a list of reaction flux limits

    The file format is detected and the file is parsed accordingly. Path can
    be given as a string or a context.
    """

    context = FilePathContext(path)

    format = resolve_format(None, context.filepath)
    if format == 'tsv':
        logger.debug('Parsing limits file {} as TSV'.format(
            context.filepath))
        with context.open('r') as f:
            for limit in parse_limits_table_file(f):
                yield limit
    elif format == 'yaml':
        logger.debug('Parsing limits file {} as YAML'.format(
            context.filepath))
        with context.open('r') as f:
            for limit in parse_limits_yaml_file(context, f):
                yield limit
    else:
        raise ParseError('Unable to detect format of limits file {}'.format(
            context.filepath))


def parse_model_group(path, group):
    """Parse a structured model group as obtained from a YAML file

    Path can be given as a string or a context.
    """

    context = FilePathContext(path)

    for reaction_id in group.get('reactions', []):
        yield convert_to_unicode(reaction_id)

    # Parse subgroups
    for reaction_id in parse_model_group_list(
            context, group.get('groups', [])):
        yield reaction_id


def parse_model_group_list(path, groups):
    """Parse a structured list of model groups as obtained from a YAML file

    Yields reaction IDs. Path can be given as a string or a context.
    """

    context = FilePathContext(path)
    for model_group in groups:
        if 'include' in model_group:
            include_context = context.resolve(model_group['include'])
            for reaction_id in parse_model_file(include_context):
                yield reaction_id
        else:
            for reaction_id in parse_model_group(context, model_group):
                yield reaction_id


def parse_model_yaml_file(path, f):
    """Parse a file as a YAML-format list of model reaction groups

    Path can be given as a string or a context.
    """
    return parse_model_group_list(path, yaml_load(f))


def parse_model_table_file(path, f):
    """Parse a file as a list of model reactions

    Yields reactions IDs. Path can be given as a string or a context.
    """

    for line in f:
        line, _, comment = convert_to_unicode(line).partition('#')
        line = line.strip()
        if line == '':
            continue

        yield line


def parse_model_file(path):
    """Parse a file as a list of model reactions

    The file format is detected and the file is parsed accordinly. The file is
    specified as a file path that will be opened for reading. Path can be given
    as a string or a context.
    """

    context = FilePathContext(path)

    format = resolve_format(None, context.filepath)
    if format == 'tsv':
        logger.debug('Parsing model file {} as TSV'.format(context.filepath))
        with context.open('r') as f:
            for reaction_id in parse_model_table_file(context, f):
                yield reaction_id
    elif format == 'yaml':
        logger.debug('Parsing model file {} as YAML'.format(context.filepath))
        with context.open('r') as f:
            for reaction_id in parse_model_yaml_file(context, f):
                yield reaction_id


# Threshold for converting reactions into dictionary representation.
_MAX_REACTION_LENGTH = 10


# Create wrapper representer for text_type for Py2/3 compatibility.
if PY3:
    def _represent_text_type(dumper, data):
        return dumper.represent_str(data)
else:
    def _represent_text_type(dumper, data):
        return dumper.represent_unicode(data)


# Define custom dict representers for YAML
# This allows reading/writing Python OrderedDicts in the correct order.
# See: https://stackoverflow.com/questions/5121931/in-python-how-can-you-load-yaml-mappings-as-ordereddicts  # noqa
def _dict_representer(dumper, data):
    return dumper.represent_dict(iteritems(data))


def _set_representer(dumper, data):
    return dumper.represent_list(iter(data))


def _boolean_expression_representer(dumper, data):
    return _represent_text_type(dumper, text_type(data))


def _reaction_representer(dumper, data):
    """Generate a parsable reaction representation to the YAML parser.

    Check the number of compounds in the reaction, if it is larger than 10,
    then transform the reaction data into a list of directories with all
    attributes in the reaction; otherwise, just return the text_type format
    of the reaction data.
    """
    if len(data.compounds) > _MAX_REACTION_LENGTH:
        def dict_make(compounds):
            for compound, value in compounds:
                yield OrderedDict([
                    ('id', text_type(compound.name)),
                    ('compartment', compound.compartment),
                    ('value', value)])

        left = list(dict_make(data.left))
        right = list(dict_make(data.right))

        direction = data.direction == Direction.Both

        reaction = OrderedDict()
        reaction['reversible'] = direction
        if data.direction == Direction.Reverse:
            reaction['left'] = right
            reaction['right'] = left
        else:
            reaction['left'] = left
            reaction['right'] = right

        return dumper.represent_data(reaction)
    else:
        return _represent_text_type(dumper, text_type(data))


def _formula_representer(dumper, data):
    return _represent_text_type(dumper, text_type(data))


def _decimal_representer(dumper, data):
    # Code from float_representer in PyYAML.
    if data % 1 == 0:
        return dumper.represent_int(int(data))
    elif math.isnan(data):
        value = '.nan'
    elif data == float('inf'):
        value = '.inf'
    elif data == float('-inf'):
        value = '-.inf'
    else:
        value = text_type(data).lower()
        if '.' not in value and 'e' in value:
            value = value.replace('e', '.0e', 1)
    return dumper.represent_scalar('tag:yaml.org,2002:float', value)


class ModelWriter(object):
    """Writer for native (YAML) format."""

    def __init__(self):
        self._yaml_args = {
            'default_flow_style': False,
            'encoding': 'utf-8',
            'allow_unicode': True,
            'width': 79
        }

    def _dump(self, stream, data):
        if hasattr(yaml, 'CSafeDumper'):
            dumper = yaml.CSafeDumper(stream, **self._yaml_args)
        else:
            dumper = yaml.SafeDumper(stream, **self._yaml_args)

        dumper.add_representer(OrderedDict, _dict_representer)
        dumper.add_representer(set, _set_representer)
        dumper.add_representer(frozenset, _set_representer)
        dumper.add_representer(
            boolean.Expression, _boolean_expression_representer)
        dumper.add_representer(Reaction, _reaction_representer)
        dumper.add_representer(Formula, _formula_representer)
        dumper.add_representer(Decimal, _decimal_representer)

        dumper.ignore_aliases = lambda *args: True

        try:
            dumper.open()
            dumper.represent(data)
            dumper.close()
        finally:
            dumper.dispose()

    def convert_compartment_entry(self, compartment, adjacencies):
        """Convert compartment entry to YAML dict.

        Args:
            compartment: :class:`psamm.datasource.entry.CompartmentEntry`.
            adjacencies: Sequence of IDs or a single ID of adjacent
                compartments (or None).
        """
        d = OrderedDict()
        d['id'] = compartment.id
        if adjacencies is not None:
            d['adjacent_to'] = adjacencies

        order = {key: i for i, key in enumerate(['name'])}
        prop_keys = set(compartment.properties)
        for prop in sorted(prop_keys,
                           key=lambda x: (order.get(x, 1000), x)):
            if compartment.properties[prop] is not None:
                d[prop] = compartment.properties[prop]

        return d

    def convert_compound_entry(self, compound):
        """Convert compound entry to YAML dict."""
        d = OrderedDict()
        d['id'] = compound.id

        order = {
            key: i for i, key in enumerate(
                ['name', 'formula', 'formula_neutral', 'charge', 'kegg',
                 'cas'])}
        prop_keys = (
            set(compound.properties) - {'boundary', 'compartment'})
        for prop in sorted(prop_keys,
                           key=lambda x: (order.get(x, 1000), x)):
            if compound.properties[prop] is not None:
                d[prop] = compound.properties[prop]

        return d

    def convert_reaction_entry(self, reaction):
        """Convert reaction entry to YAML dict."""
        d = OrderedDict()
        d['id'] = reaction.id

        def is_equation_valid(equation):
            # If the equation is a Reaction object, it must have non-zero
            # number of compounds.
            return (equation is not None and (
                    not isinstance(equation, Reaction) or
                    len(equation.compounds) > 0))

        order = {
            key: i for i, key in enumerate(
                ['name', 'genes', 'equation', 'subsystem', 'ec'])}
        prop_keys = (set(reaction.properties) -
                     {'lower_flux', 'upper_flux', 'reversible'})
        for prop in sorted(prop_keys, key=lambda x: (order.get(x, 1000), x)):
            if reaction.properties[prop] is None:
                continue
            d[prop] = reaction.properties[prop]
            if prop == 'equation' and not is_equation_valid(d[prop]):
                del d[prop]

        return d

    def _write_entries(self, stream, entries, converter, properties=None):
        """Write iterable of entries as YAML object to stream.

        Args:
            stream: File-like object.
            entries: Iterable of entries.
            converter: Conversion function from entry to YAML object.
            properties: Set of compartment properties to output (or None to
                output all).
        """
        def iter_entries():
            for c in entries:
                entry = converter(c)
                if entry is None:
                    continue
                if properties is not None:
                    entry = OrderedDict(
                        (key, value) for key, value in iteritems(entry)
                        if key == 'id' or key in properties)
                yield entry

        self._dump(stream, list(iter_entries()))

    def write_compartments(self, stream, compartments, adjacencies,
                           properties=None):
        """Write iterable of compartments as YAML object to stream.

        Args:
            stream: File-like object.
            compartments: Iterable of compartment entries.
            adjacencies: Dictionary mapping IDs to adjacent compartment IDs.
            properties: Set of compartment properties to output (or None to
                output all).
        """
        def convert(entry):
            return self.convert_compartment_entry(
                entry, adjacencies.get(entry.id))

        self._write_entries(stream, compartments, convert, properties)

    def write_compounds(self, stream, compounds, properties=None):
        """Write iterable of compounds as YAML object to stream.

        Args:
            stream: File-like object.
            compounds: Iterable of compound entries.
            properties: Set of compound properties to output (or None to output
                all).
        """
        self._write_entries(
            stream, compounds, self.convert_compound_entry, properties)

    def write_reactions(self, stream, reactions, properties=None):
        """Write iterable of reactions as YAML object to stream.

        Args:
            stream: File-like object.
            compounds: Iterable of reaction entries.
            properties: Set of reaction properties to output (or None to output
                all).
        """
        self._write_entries(
            stream, reactions, self.convert_reaction_entry, properties)


def reaction_signature(eq, direction=False, stoichiometry=False):
    """Return unique signature object for :class:`Reaction`.

    Signature objects are hashable, and compare equal only if the reactions
    are considered the same according to the specified rules.

    Args:
        direction: Include reaction directionality when considering equality.
        stoichiometry: Include stoichiometry when considering equality.
    """
    def compounds_sig(compounds):
        if stoichiometry:
            return tuple(sorted(compounds))
        else:
            return tuple(sorted(compound for compound, _ in compounds))

    left = compounds_sig(eq.left)
    right = compounds_sig(eq.right)

    if left < right:
        reaction_sig = left, right
        direction_sig = eq.direction
    else:
        reaction_sig = right, left
        direction_sig = eq.direction.flipped()

    if direction:
        return reaction_sig, direction_sig
    return reaction_sig



# This file is part of PSAMM.
#
# PSAMM is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PSAMM is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PSAMM.  If not, see <http://www.gnu.org/licenses/>.
#
# Copyright 2014-2015  Jon Lund Steffensen <jon_steffensen@uri.edu>
# Copyright 2015-2020  Keith Dufault-Thompson <keitht547@my.uri.edu>

from __future__ import unicode_literals

import logging

from ..command import Command, FilePrefixAppendAction, convert_to_unicode
from ..formula import Formula
from ..balancecheck import formula_balance

logger = logging.getLogger(__name__)


class FormulaBalanceCommand(Command):
    """Check whether reactions in the model are elementally balanced.

    Balanced reactions are those reactions where the number of elements
    (atoms) is consistent on the left and right side of the reaction equation.
    Reactions that are not balanced will be printed out.
    """

    @classmethod
    def init_parser(cls, parser):
        parser.add_argument(
            '--exclude', metavar='reaction', action=FilePrefixAppendAction,
            type=convert_to_unicode, default=[],
            help='Exclude reaction from balance check')
        super(FormulaBalanceCommand, cls).init_parser(parser)

    def run(self):
        """Run formula balance command"""

        # Create a set of excluded reactions
        exclude = set(self._args.exclude)
        count = 0
        unbalanced = 0
        unchecked = 0
        for reaction, result in formula_balance(self._model):
            count += 1

            if reaction.id in exclude or reaction.equation is None:
                continue

            if result is None:
                unchecked += 1
                continue

            left_form, right_form = result
            if right_form != left_form:
                unbalanced += 1
                right_missing, left_missing = Formula.balance(
                    right_form, left_form)

                print('{}\t{}\t{}\t{}\t{}'.format(
                    reaction.id, left_form, right_form,
                    left_missing, right_missing))

        logger.info('Unbalanced reactions: {}/{}'.format(unbalanced, count))
        logger.info('Unchecked reactions due to missing formula: {}/{}'.format(
            unchecked, count))
        logger.info('Reactions excluded from check: {}/{}'.format(
            len(exclude), count))



# This file is part of PSAMM.
#
# PSAMM is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PSAMM is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PSAMM.  If not, see <http://www.gnu.org/licenses/>.
#
# Copyright 2014-2017  Jon Lund Steffensen <jon_steffensen@uri.edu>
# Copyright 2015-2020  Keith Dufault-Thompson <keitht547@my.uri.edu>

from __future__ import unicode_literals

import time
import logging

from ..command import (LoopRemovalMixin, ObjectiveMixin, SolverCommandMixin,
                       MetabolicMixin, Command)
from .. import fluxanalysis

logger = logging.getLogger(__name__)


class FluxBalanceCommand(MetabolicMixin, LoopRemovalMixin, ObjectiveMixin,
                         SolverCommandMixin, Command):
    """Run flux balance analysis on the model."""

    _supported_loop_removal = ['none', 'tfba', 'l1min']

    @classmethod
    def init_parser(cls, parser):
        parser.add_argument(
            '--all-reactions', help='Show all reaction fluxes',
            action='store_true')
        parser.add_argument(
            '--epsilon', type=float,
            help='Threshold for non-zero reaction fluxes', default=1e-5)
        super(FluxBalanceCommand, cls).init_parser(parser)

    def run(self):
        """Run flux analysis command."""

        # Load compound information
        def compound_name(id):
            if id not in self._model.compounds:
                return id
            return self._model.compounds[id].properties.get('name', id)

        # Reaction genes information
        def reaction_genes_string(id):
            if id not in self._model.reactions:
                return ''
            return self._model.reactions[id].properties.get('genes', '')

        reaction = self._get_objective()
        if not self._mm.has_reaction(reaction):
            self.fail(
                'Specified reaction is not in model: {}'.format(reaction))

        loop_removal = self._get_loop_removal_option()
        if loop_removal == 'none':
            result = self.run_fba(reaction)
        elif loop_removal == 'l1min':
            result = self.run_fba_minimized(reaction)
        elif loop_removal == 'tfba':
            result = self.run_tfba(reaction)

        optimum = None
        total_reactions = 0
        nonzero_reactions = 0
        for reaction_id, flux in sorted(result):
            total_reactions += 1
            if abs(flux) > self._args.epsilon:
                nonzero_reactions += 1

            if abs(flux) > self._args.epsilon or self._args.all_reactions:
                rx = self._mm.get_reaction(reaction_id)
                rx_trans = rx.translated_compounds(compound_name)
                genes = reaction_genes_string(reaction_id)
                print('{}\t{}\t{}\t{}'.format(
                    reaction_id, flux, rx_trans, genes))

            # Remember flux of requested reaction
            if reaction_id == reaction:
                optimum = flux

        logger.info('Objective flux: {}'.format(optimum))
        logger.info('Reactions at zero flux: {}/{}'.format(
            total_reactions - nonzero_reactions, total_reactions))

    def run_fba(self, reaction):
        """Run standard FBA on model."""
        solver = self._get_solver()
        p = fluxanalysis.FluxBalanceProblem(self._mm, solver)

        start_time = time.time()

        try:
            p.maximize(reaction)
        except fluxanalysis.FluxBalanceError as e:
            self.report_flux_balance_error(e)

        logger.info('Solving took {:.2f} seconds'.format(
            time.time() - start_time))

        for reaction_id in self._mm.reactions:
            yield reaction_id, p.get_flux(reaction_id)

    def run_fba_minimized(self, reaction):
        """Run normal FBA and flux minimization on model."""

        epsilon = self._args.epsilon
        solver = self._get_solver()

        p = fluxanalysis.FluxBalanceProblem(self._mm, solver)

        start_time = time.time()

        # Maximize reaction flux
        try:
            p.maximize(reaction)
        except fluxanalysis.FluxBalanceError as e:
            self.report_flux_balance_error(e)

        fluxes = {r: p.get_flux(r) for r in self._mm.reactions}

        # Run flux minimization
        flux_var = p.get_flux_var(reaction)
        p.prob.add_linear_constraints(flux_var == p.get_flux(reaction))
        p.minimize_l1()

        logger.info('Solving took {:.2f} seconds'.format(
            time.time() - start_time))

        count = 0
        for reaction_id in self._mm.reactions:
            flux = p.get_flux(reaction_id)
            if abs(flux - fluxes[reaction_id]) > epsilon:
                count += 1
            yield reaction_id, flux
        logger.info('Minimized reactions: {}'.format(count))

    def run_tfba(self, reaction):
        """Run FBA and tFBA on model."""

        solver = self._get_solver(integer=True)
        p = fluxanalysis.FluxBalanceProblem(self._mm, solver)

        start_time = time.time()

        p.add_thermodynamic()

        try:
            p.maximize(reaction)
        except fluxanalysis.FluxBalanceError as e:
            self.report_flux_balance_error(e)

        logger.info('Solving took {:.2f} seconds'.format(
            time.time() - start_time))

        for reaction_id in self._mm.reactions:
            yield reaction_id, p.get_flux(reaction_id)



# This file is part of PSAMM.
#
# PSAMM is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PSAMM is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PSAMM.  If not, see <http://www.gnu.org/licenses/>.
#
# Copyright 2014-2017  Jon Lund Steffensen <jon_steffensen@uri.edu>
# Copyright 2015-2020  Keith Dufault-Thompson <keitht547@my.uri.edu>
# Copyright 2016  Chao Liu <lcddzyx@gmail.com>

import os
import unittest
import tempfile
import shutil

from psamm.datasource.native import ModelReader
from psamm.metabolicmodel import MetabolicModel
from psamm.database import DictDatabase
from psamm.datasource.reaction import parse_reaction
from psamm.expression import boolean
from psamm.lpsolver import generic
from psamm import fluxanalysis, randomsparse


class TestGetGeneAssociation(unittest.TestCase):
    def setUp(self):
        self._model_dir = tempfile.mkdtemp()
        with open(os.path.join(self._model_dir, 'model.yaml'), 'w') as f:
            f.write('\n'.join([
                '---',
                'reactions:',
                '  - id: rxn_1',
                '    equation: A[e] <=> B[c]',
                '    genes: gene_1 and gene_2',
                '  - id: rxn_2',
                '    equation: A[e] => C[c]',
                '    genes: gene_3',
                '  - id: rxn_3',
                '    equation: A[e] => D[e]',
                '  - id: rxn_4',
                '    equation: C[c] => D[e]',
                '    genes: [gene_4, gene_5, gene_6]',
                'compounds:',
                '  - id: A',
                '  - id: B',
                '  - id: C',
                '  - id: D',
                'exchange:',
                '  - compartment: e',
                '    compounds:',
                '      - id: A',
                '        reaction: rxn_5',
                '      - id: D',
                '        reaction: rxn_6',
                'model:',
                '  - reactions:',
                '     - rxn_3',
                '     - rxn_4',
            ]))
        self._model = ModelReader.reader_from_path(
            self._model_dir).create_model()

    def tearDown(self):
        shutil.rmtree(self._model_dir)

    def test_get_gene_association(self):
        expected_association = {
            'rxn_1': boolean.Expression('gene_1 and gene_2'),
            'rxn_2': boolean.Expression('gene_3'),
            'rxn_4': boolean.Expression('gene_4 and gene_6 and gene_5')
        }

        self.assertEqual(
            dict(randomsparse.get_gene_associations(self._model)),
            expected_association)


class TestGetExchangeReactions(unittest.TestCase):
    def setUp(self):
        self._database = DictDatabase()
        self._database.set_reaction('rxn_1', parse_reaction('|A| <=>'))
        self._database.set_reaction('rxn_2', parse_reaction('|A| <=> |B|'))
        self._database.set_reaction('rxn_3', parse_reaction('|C| <=> |B|'))
        self._database.set_reaction('rxn_4', parse_reaction('|C| <=>'))
        self._mm = MetabolicModel.load_model(
            self._database, self._database.reactions)

    def test_get_exchange_reactions(self):
        expected_reactions = {'rxn_1', 'rxn_4'}
        self.assertEqual(
            set(randomsparse.get_exchange_reactions(self._mm)),
            expected_reactions)


class TestReactionDeletionStrategy(unittest.TestCase):
    def setUp(self):
        self._database = DictDatabase()
        self._database.set_reaction('rxn_1', parse_reaction('|A| <=>'))
        self._database.set_reaction('rxn_2', parse_reaction('|A| <=> |B|'))
        self._database.set_reaction('rxn_3', parse_reaction('|C| <=> |B|'))
        self._database.set_reaction('rxn_4', parse_reaction('|C| <=>'))
        self._mm = MetabolicModel.load_model(
            self._database, self._database.reactions)

        self._strategy = randomsparse.ReactionDeletionStrategy(self._mm)

    def test_method_get_all(self):
        expected_total = {'rxn_1', 'rxn_2', 'rxn_3', 'rxn_4'}
        self.assertEqual(set(self._strategy.entities), expected_total)

    def test_method_tests(self):
        expected_reactions = {
            'rxn_1': {'rxn_1'},
            'rxn_2': {'rxn_2'},
            'rxn_3': {'rxn_3'},
            'rxn_4': {'rxn_4'}
        }
        self.assertEqual(dict(self._strategy.iter_tests()), expected_reactions)


class TestGeneDeletionStrategy(unittest.TestCase):
    def setUp(self):
        self._database = DictDatabase()
        self._database.set_reaction('rxn_1', parse_reaction('|A| <=>'))
        self._database.set_reaction('rxn_2', parse_reaction('|A| <=> |B|'))
        self._database.set_reaction('rxn_3', parse_reaction('|C| <=> |B|'))
        self._database.set_reaction('rxn_4', parse_reaction('|C| <=>'))
        self._mm = MetabolicModel.load_model(
            self._database, self._database.reactions)
        self._assoc = {'rxn_2': boolean.Expression('gene_1 or gene_2')}

        self._strategy = randomsparse.GeneDeletionStrategy(
            self._mm, self._assoc)

    def test_method_get_all(self):
        expected_total = {'gene_1', 'gene_2'}
        self.assertEqual(set(self._strategy.entities), expected_total)

    def test_method_tests_and_delete(self):
        expected_genes = {'gene_1', 'gene_2'}
        expected_reaction_set = {'rxn_2'}
        test_dict = {}

        for i, (gene, deleted_reac) in enumerate(self._strategy.iter_tests()):
            self._strategy.delete(gene, deleted_reac)
            if i == 0:
                self.assertEqual(deleted_reac, set())
            else:
                self.assertEqual(deleted_reac, {'rxn_2'})
            test_dict[gene] = deleted_reac

        self.assertTrue(all(x in test_dict for x in expected_genes))
        self.assertTrue(expected_reaction_set in test_dict.values())


class TestRandomSparse(unittest.TestCase):
    def setUp(self):
        self._database = DictDatabase()
        self._database.set_reaction('rxn_1', parse_reaction('A[e] => B[e]'))
        self._database.set_reaction('rxn_2', parse_reaction('B[e] => C[e]'))
        self._database.set_reaction('rxn_3', parse_reaction('B[e] => D[e]'))
        self._database.set_reaction('rxn_4', parse_reaction('C[e] => E[e]'))
        self._database.set_reaction('rxn_5', parse_reaction('D[e] => E[e]'))
        self._database.set_reaction('rxn_6', parse_reaction('E[e] =>'))
        self._database.set_reaction('ex_A', parse_reaction('A[e] <=>'))

        self._mm = MetabolicModel.load_model(
            self._database, self._database.reactions)
        self._assoc = {
            'rxn_1': boolean.Expression('gene_1'),
            'rxn_2': boolean.Expression('gene_2 or gene_3'),
            'rxn_5': boolean.Expression('gene_3 and gene_4')
        }
        self._obj_reaction = 'rxn_6'

        try:
            self._solver = generic.Solver()
        except generic.RequirementsError:
            self.skipTest('Unable to find an LP solver for tests')

        self._prob = fluxanalysis.FluxBalanceProblem(self._mm, self._solver)

    def test_random_sparse_reaction_strategy(self):
        expected = [
            ({'rxn_1', 'rxn_6', 'ex_A', 'rxn_2', 'rxn_4'}, {'rxn_3', 'rxn_5'}),
            ({'rxn_1', 'rxn_6', 'ex_A', 'rxn_3', 'rxn_5'}, {'rxn_2', 'rxn_4'})
        ]

        strategy = randomsparse.ReactionDeletionStrategy(self._mm)
        essential, deleted = randomsparse.random_sparse(
            strategy,
            self._prob,
            self._obj_reaction,
            flux_threshold=100)

        self.assertTrue((essential, deleted) in expected)

    def test_random_sparse_gene_strategy(self):
        expected = [
            ({'gene_1', 'gene_2'}, {'gene_3', 'gene_4'}),
            ({'gene_1', 'gene_3'}, {'gene_2', 'gene_4'})
        ]
        strategy = randomsparse.GeneDeletionStrategy(
            self._mm, self._assoc)
        essential, deleted = randomsparse.random_sparse(
            strategy,
            self._prob,
            self._obj_reaction,
            flux_threshold=100)

        self.assertTrue((essential, deleted) in expected)


if __name__ == '__main__':
    unittest.main()



# This file is part of PSAMM.
#
# PSAMM is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PSAMM is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PSAMM.  If not, see <http://www.gnu.org/licenses/>.
#
# Copyright 2016-2017  Jon Lund Steffensen <jon_steffensen@uri.edu>
# Copyright 2015-2020  Keith Dufault-Thompson <keitht547@my.uri.edu>

"""Functions for predicting compound pairs using the MapMaker algorithm.

The MapMaker algorithm is described in [Tervo16]_.
"""

from itertools import product

from six import iteritems

from .lpsolver import lp
from .formula import Atom, Radical, Formula


def default_weight(element):
    """Return weight of formula element.

    This implements the default weight proposed for MapMaker.
    """
    if element in (Atom.N, Atom.O, Atom.P):
        return 0.4
    elif isinstance(element, Radical):
        return 40.0
    return 1.0


def _weighted_formula(form, weight_func):
    """Yield weight of each formula element."""
    for e, mf in form.items():
        if e == Atom.H:
            continue

        yield e, mf, weight_func(e)


def _transfers(reaction, delta, elements, result, epsilon):
    """Yield transfers obtained from result."""
    left = set(c for c, _ in reaction.left)
    right = set(c for c, _ in reaction.right)
    for c1, c2 in product(left, right):
        items = {}
        for e in elements:
            v = result.get_value(delta[c1, c2, e])
            nearest_int = round(v)
            if abs(v - nearest_int) < epsilon:
                v = int(nearest_int)
            if v >= epsilon:
                items[e] = v

        if len(items) > 0:
            yield (c1, c2), Formula(items)


def _reaction_to_dicts(reaction):
    """Convert a reaction to reduced left, right dictionaries."""

    def dict_from_iter_sum(it):
        d = {}
        for k, v in it:
            if k not in d:
                d[k] = 0
            d[k] += v
        return d

    left = dict_from_iter_sum(reaction.left)
    right = dict_from_iter_sum(reaction.right)

    return left, right


class UnbalancedReactionError(ValueError):
    """Raised when an unbalanced reaction is provided."""


def predict_compound_pairs(reaction, compound_formula, solver, epsilon=1e-5,
                           alt_elements=None, weight_func=default_weight):
    """Predict compound pairs of reaction using MapMaker.

    Yields all solutions as dictionaries with compound pairs as keys and
    formula objects as values.

    Args:
        reaction: :class:`psamm.reaction.Reaction` object.
        compound_formula: Dictionary mapping compound IDs to formulas. Formulas
            must be flattened.
        solver: LP solver (MILP).
        epsilon: Threshold for rounding floating-point values to integers in
            the predicted transfers.
        alt_elements: Iterable of elements to consider for alternative
              solutions. Only alternate solutions that have different transfers
              of these elements will be returned (default=all elements).
        weight_func: Optional function that returns a weight for a formula
            element (should handle specific Atom and Radical objects). By
            default, the standard MapMaker weights will be used
            (H=0, R=40, N=0.4, O=0.4, P=0.4, *=1).
    """
    elements = set()
    for compound, value in reaction.compounds:
        if compound.name not in compound_formula:
            return

        f = compound_formula[compound.name]
        elements.update(e for e, _, _ in _weighted_formula(f, weight_func))

    if len(elements) == 0:
        return

    p = solver.create_problem()

    x = p.namespace(name='x')
    y = p.namespace(name='y')
    m = p.namespace(name='m')
    q = p.namespace(name='q')
    omega = p.namespace(name='omega')
    gamma = p.namespace(name='gamma')
    delta = p.namespace(name='delta')

    left, right = _reaction_to_dicts(reaction)

    objective = 0
    for c1, c2 in product(left, right):
        m.define([(c1, c2)], lower=0)
        q.define([(c1, c2)], lower=0)
        omega.define([(c1, c2)], types=lp.VariableType.Binary)

        objective += 100 * m[c1, c2] + 90 * q[c1, c2] - 0.1 * omega[c1, c2]

        gamma.define(((c1, c2, e) for e in elements),
                     types=lp.VariableType.Binary)
        delta.define(((c1, c2, e) for e in elements), lower=0)

        # Eq 12
        x.define([(c1, c2)], types=lp.VariableType.Binary)
        y.define([(c1, c2)], types=lp.VariableType.Binary)
        p.add_linear_constraints(y[c1, c2] <= 1 - x[c1, c2])

        # Eq 6, 9
        delta_wsum = delta.expr(
            ((c1, c2, e), weight_func(e)) for e in elements)
        p.add_linear_constraints(
            m[c1, c2] <= delta_wsum, q[c1, c2] <= delta_wsum)

        objective -= gamma.sum((c1, c2, e) for e in elements)

    p.set_objective(objective)

    # Eq 3
    zs = {}
    for c1, v in iteritems(left):
        f = compound_formula[c1.name]
        for e in elements:
            mf = f.get(e, 0)
            delta_sum = delta.sum((c1, c2, e) for c2 in right)
            try:
                p.add_linear_constraints(delta_sum == v * mf)
            except ValueError:
                raise UnbalancedReactionError('Unable to add constraint')

        # Eq 8, 11
        x_sum = x.sum((c1, c2) for c2 in right)
        y_sum = y.sum((c1, c2) for c2 in right)
        p.add_linear_constraints(x_sum <= 1, y_sum <= 1)

        # Eq 13
        zs[c1] = 0
        for e, mf, w in _weighted_formula(f, weight_func):
            zs[c1] += w * mf

    # Eq 2
    for c2, v in iteritems(right):
        f = compound_formula[c2.name]
        for e in elements:
            mf = f.get(e, 0)
            delta_sum = delta.sum((c1, c2, e) for c1 in left)
            try:
                p.add_linear_constraints(delta_sum == v * mf)
            except ValueError:
                raise UnbalancedReactionError('Unable to add constraint')

    for c1, v1 in iteritems(left):
        for c2 in right:
            f1 = compound_formula[c1.name]
            f2 = compound_formula[c2.name]
            for e in elements:
                mf = f1.get(e, 0)

                # Eq 4
                p.add_linear_constraints(
                    delta[c1, c2, e] <= float(v1) * mf * omega[c1, c2])

                # Eq 5
                if e in f2:
                    p.add_linear_constraints(
                        delta[c1, c2, e] <= float(v1) * mf * gamma[c1, c2, e])

            # Eq 7, 10
            p.add_linear_constraints(
                m[c1, c2] <= float(v1) * zs[c1] * x[c1, c2],
                q[c1, c2] <= float(v1) * zs[c1] * y[c1, c2])

    try:
        result = p.solve(lp.ObjectiveSense.Maximize)
    except lp.SolverError:
        raise UnbalancedReactionError('Unable to solve')

    first_objective = result.get_value(objective)

    # Yield the first result
    yield dict(_transfers(reaction, delta, elements, result, epsilon))

    # Find alternative solutions
    if alt_elements is None:
        alt_elements = elements
    else:
        alt_elements = elements.intersection(alt_elements)

    if len(alt_elements) == 0:
        return

    # Add a constraint that disallows the current transfer solution until no
    # alternative solutions are left.
    add_objective_constraint = True
    while True:
        elem_vars = 0
        elem_count = 0
        for c1, c2 in product(left, right):
            for e in alt_elements:
                if result.get_value(gamma[c1, c2, e]) > 0.5:
                    elem_count += 1
                    elem_vars += gamma[c1, c2, e]

        if elem_count == 0:
            break

        if add_objective_constraint:
            p.add_linear_constraints(objective >= first_objective)
            add_objective_constraint = False

        p.add_linear_constraints(elem_vars <= elem_count - 1)
        try:
            result = p.solve(lp.ObjectiveSense.Maximize)
        except lp.SolverError:
            break

        yield dict(_transfers(reaction, delta, elements, result, epsilon))



# This file is part of PSAMM.
#
# PSAMM is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PSAMM is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PSAMM.  If not, see <http://www.gnu.org/licenses/>.
#
# Copyright 2014-2017  Jon Lund Steffensen <jon_steffensen@uri.edu>
# Copyright 2015-2020  Keith Dufault-Thompson <keitht547@my.uri.edu>

from __future__ import unicode_literals

import argparse
import logging

from ..command import Command, MetabolicMixin, SolverCommandMixin
from ..fastgapfill import fastgapfill
from ..gapfilling import create_extended_model

logger = logging.getLogger(__name__)


class FastGapFillCommand(MetabolicMixin, SolverCommandMixin, Command):
    """Run the FastGapFill gap-filling algorithm on model."""

    @classmethod
    def init_parser(cls, parser):
        parser.add_argument(
            '--subset', metavar='file', type=argparse.FileType('r'),
            help='specify core reaction subset to use')
        parser.add_argument(
            '--penalty', metavar='file', type=argparse.FileType('r'),
            help='List of penalty scores for database reactions')
        parser.add_argument(
            '--db-penalty', metavar='penalty', type=float,
            help='Default penalty for database reactions')
        parser.add_argument(
            '--tp-penalty', metavar='penalty', type=float,
            help='Default penalty for transport reactions')
        parser.add_argument(
            '--ex-penalty', metavar='penalty', type=float,
            help='Default penalty for exchange reactions')
        parser.add_argument(
            '--epsilon', type=float, help='Threshold for Fastcore',
            default=1e-5)
        super(FastGapFillCommand, cls).init_parser(parser)

    def run(self):
        """Run FastGapFill command"""

        # Create solver
        solver = self._get_solver()

        # Load compound information
        def compound_name(id):
            if id not in self._model.compounds:
                return id
            return self._model.compounds[id].properties.get('name', id)

        # TODO: The exchange and transport reactions have tuple names. This
        # means that in Python 3 the reactions can no longer be directly
        # compared (e.g. while sorting) so define this helper function as a
        # workaround.
        def reaction_key(r):
            return r if isinstance(r, tuple) else (r,)

        # Calculate penalty if penalty file exists
        penalties = {}
        for id in sorted(self._mm.reactions):
            penalties[id] = 0
        if self._args.penalty is not None:
            for line in self._args.penalty:
                line, _, comment = line.partition('#')
                line = line.strip()
                if line == '':
                    continue
                rxnid, penalty = line.split(None, 1)
                penalties[rxnid] = float(penalty)

        model_extended, weights = create_extended_model(
            self._model,
            db_penalty=self._args.db_penalty,
            ex_penalty=self._args.ex_penalty,
            tp_penalty=self._args.tp_penalty,
            penalties=penalties)

        epsilon = self._args.epsilon
        core = set()
        if self._args.subset is None:
            for r in self._mm.reactions:
                if not self._mm.is_exchange(r):
                    core.add(r)
        else:
            for line in self._args.subset:
                line = line.strip()
                if line == '':
                    continue
                core.add(line)
        induced = fastgapfill(model_extended, core, weights=weights,
                              epsilon=epsilon, solver=solver)

        for reaction_id in sorted(self._mm.reactions):
            rx = self._mm.get_reaction(reaction_id)
            rxt = rx.translated_compounds(compound_name)
            print('{}\t{}\t{}\t{}'.format(
                reaction_id, 'Model', weights[reaction_id], rxt))

        for rxnid in sorted(induced, key=reaction_key):
            if self._mm.has_reaction(rxnid):
                continue
            rx = model_extended.get_reaction(rxnid)
            rxt = rx.translated_compounds(compound_name)
            print('{}\t{}\t{}\t{}'.format(
                rxnid, 'Add', weights.get(rxnid), rxt))



# This file is part of PSAMM.
#
# PSAMM is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PSAMM is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PSAMM.  If not, see <http://www.gnu.org/licenses/>.
#
# Copyright 2021-2022  Jason Vailionis <jason_vailionis@uri.edu>

from psamm.datasource.reaction import Reaction, Compound, Direction
from psamm.datasource import native
from collections import Counter
from six import add_metaclass
from pkg_resources import resource_filename
import logging
import sys
import os
import abc
import yaml
import pandas as pd
import numpy as np
if sys.version_info.minor > 5:
    try:
        from Bio import SeqIO
        from Bio import Seq
    except ImportError:
        quit("No biopython package found. "
             "Please run <pip install biopython>")

logger = logging.getLogger(__name__)


@add_metaclass(abc.ABCMeta)
class Command(object):
    """Represents a command in the interface, operating on a model.

    The constructor will be given the NativeModel and the command line
    namespace. The subclass must implement :meth:`run` to handle command
    execution. The doc string will be used as documentation for the command
    in the command line interface.

    In addition, :meth:`init_parser` can be implemented as a classmethod which
    will allow the command to initialize an instance of
    :class:`argparse.ArgumentParser` as desired. The resulting argument
    namespace will be passed to the constructor.
    """

    def __init__(self, args):
        self._args = args

    @classmethod
    def init_parser(cls, parser):
        """Initialize command line parser (:class:`argparse.ArgumentParser`)"""

    @abc.abstractmethod
    def run(self):
        """Execute command"""

    def argument_error(self, msg):
        """Raise error indicating error parsing an argument."""
        raise CommandError(msg)

    def fail(self, msg, exc=None):
        """Exit command as a result of a failure."""
        logger.error(msg)
        if exc is not None:
            logger.debug("Command failure caused by exception!", exc_info=exc)
        sys.exit(1)


class InputError(Exception):
    """Exception used to signal a general input error."""


class CommandError(Exception):
    """Error from running a command.

    This should be raised from a ``Command.run()`` if any arguments are
    misspecified. When the command is run and the ``CommandError`` is raised,
    the caller will exit with an error code and print appropriate usage
    information.
    """


# Returns df with additional columns for stoichiometry
def calc_stoichiometry_df(df, counts):
    df["counts"] = pd.Series(counts)
    df["composition"] = df.counts / df.counts.sum()
    df["mass"] = df.mol_weight*df.composition
    df["stoichiometry"] = df.composition / df.mass.sum()
    df = df.fillna(0)
    return df


def load_compound_data():
    df = pd.read_csv(resource_filename("psamm",
                     "external-data/biomass_compound_descriptions.tsv"),
                     sep="\t", na_values=["None"])
    df.set_index(df.id.values, inplace=True)
    return df


def make_template(compounds_df):
    template = compounds_df[["id", "name"]]
    template["custom_id"] = np.nan
    return template


def apply_config(df, config_path):
    config = pd.read_csv(config_path)
    save_index = df.index
    df = pd.merge(df, config, how="left", on=["id", "name"])
    df.set_index(save_index, inplace=True)
    df.custom_id = df.custom_id.fillna(df.id)
    df.id = df.custom_id
    return df

# return model_dict


def load_model(model_path):
    with open(model_path, "r") as f:
        model_dict = yaml.safe_load(f)
    return model_dict


def check_missing_cpds(model_path, df, config, yaml_args):
    # Check if needed compounds are in the model, add them if not
    model_dir = os.path.dirname(model_path)
    model_reader = native.ModelReader.reader_from_path(model_path)
    compounds = [r.id for r in model_reader.parse_compounds()]
    if len(compounds) > 0:
        missing_cpds = df.query("id not in @compounds")
        droplist = ["type", "code"]
        if config:
            droplist.append("custom_id")
        missing_cpds = missing_cpds.drop(columns=droplist)
        if len(missing_cpds) > 0:
            logger.warning("\033[1mThe following compounds are required for "
                           "biomass equations but are missing from the model: "
                           "\033[0m{}\n\033[1mPredefined versions of these "
                           "compounds will be generated in {}. To use custom "
                           "biomass compound IDs, use --config\033[0m"
                           "".format(list(missing_cpds.id),
                                     os.path.join(model_dir,
                                                  "biomass_compounds.yaml")))
            missing_cpd_entries = [v.dropna().to_dict()
                                   for k, v in missing_cpds.iterrows()]
            with open(os.path.join(model_dir,
                                   "biomass_compounds.yaml"), "a+") as f:
                yaml.dump(missing_cpd_entries, f, **yaml_args)


def fix_cell_compartment(model_dict):
    try:
        cell_compartment = model_dict["default_compartment"]
    except KeyError:
        cell_compartment = "c"
        model_dict["default_compartment"] = "c"
    if cell_compartment is None:
        cell_compartment = "c"
        model_dict["default_compartment"] = "c"
    return cell_compartment


def update_model_yaml(model_path, model_dict, cell_compartment,
                      biomass_name, yaml_args):
    exists = False  # Checking if biomass_reactions.yaml already included
    for file in model_dict["reactions"]:
        if "biomass_reactions.yaml" in file["include"]:
            exists = True
    if not exists:
        model_dict["reactions"].append({"include":
                                        "./biomass_reactions.yaml"})

    exists = False  # Checking if biomass_compounds.yaml already included
    for file in model_dict["compounds"]:
        if "biomass_compounds.yaml" in file["include"]:
            exists = True
    if not exists:
        model_dict["compounds"].append({"include":
                                        "./biomass_compounds.yaml"})

    model_dict["biomass"] = biomass_name

    os.rename(model_path, model_path + ".tmp")
    with open(model_path, "w") as f:
        yaml.dump(model_dict, f, **yaml_args)
    os.remove(model_path + ".tmp")


def count_DNA(genome, df, cell_compartment, decimals=6):
    DNA = df[df.type == "DNA"].copy().set_index("code")
    DNA_counts = Counter()
    for seq in genome:
        DNA_counts += Counter(genome[seq].seq)

    calc_stoichiometry_df(DNA, DNA_counts)
    total = DNA.stoichiometry.sum()
    ids_fwd = list(DNA.id) + list(df.loc[["C00002", "C00001"]].id.values)
    stoich_fwd = list(DNA.stoichiometry) + [2*total, 2*total]
    compound_fwd = {Compound(id).in_compartment(cell_compartment):
                    round(-1*n, decimals) for id, n in zip(ids_fwd,
                                                           stoich_fwd)}
    ids_rev = list(df.loc[["dna", "C00008", "C00009", "C00013"]].id.values)
    stoich_rev = [1, 2*total, 2*total, total]
    compound_rev = {Compound(id).in_compartment(cell_compartment):
                    round(n, decimals) for id, n in zip(ids_rev, stoich_rev)}
    dna_rxn = Reaction(Direction.Forward, {**compound_fwd, **compound_rev})
    dna_rxn_entry = {"id": "dna_met",
                     "name": "DNA",
                     "equation": str(dna_rxn),
                     "pathways": ["Biomass"]}
    return dna_rxn_entry


def count_Prot(proteome, df, cell_compartment, decimals=6):
    Prot = df[df.type == "Prot"].copy().set_index("code")
    Prot_counts = Counter()
    for seq in proteome:
        Prot_counts += Counter(proteome[seq].seq)

    # Protein Reaction formation
    calc_stoichiometry_df(Prot, Prot_counts)
    total = Prot.stoichiometry.sum()
    ids_fwd = list(df[df.type == "aatrna"].id) + \
        list(df.loc[["C00044", "C00001"]].id.values)
    stoich_fwd = list(Prot.stoichiometry) + [2*total, 2*total]
    compound_fwd = {Compound(id).in_compartment(cell_compartment):
                    round(-1*n, decimals) for id, n in zip(ids_fwd,
                                                           stoich_fwd)}
    ids_rev = list(df[df.type == "trna"].id) +\
        list(df.loc[["protein", "C00035", "C00009"]].id.values)
    stoich_rev = list(Prot.stoichiometry) + [1, 2*total, 2*total]
    compound_rev = {Compound(id).in_compartment(cell_compartment):
                    round(n, decimals) for id, n in zip(ids_rev, stoich_rev)}
    prot_rxn = Reaction(Direction.Forward, {**compound_fwd, **compound_rev})
    prot_rxn_entry = {"id": "pro_met",
                      "name": "Protein",
                      "equation": str(prot_rxn),
                      "pathways": ["Biomass"]}
    return prot_rxn_entry


def count_RNA(genome, gff, df, cell_compartment, decimals=6):
    RNA = df[df.type == "RNA"].copy().set_index("code")
    RNA_counts = Counter()
    # Parsing through the gff and counting bases in coding regions to get
    # the total RNA base counts
    with open(gff, "r") as gff_file:
        for line in gff_file:
            if not line.startswith("#"):
                L = line.strip().split("\t")
                if L[2] == "CDS":
                    try:
                        seq_entry = genome[L[0]]
                    except KeyError:
                        logger.warning("The annotation {s} does not match "
                                       "any sequence names in genome file, "
                                       "ignoring...".format(s=L[0]))
                    # FIX STRANDEDNESS: if - use reverse complement
                    if L[6] == "-":
                        RNA_counts += Counter(Seq.complement(
                            seq_entry.seq[int(L[3]):int(L[4])]))
                    else:
                        RNA_counts += Counter(
                            seq_entry.seq[int(L[3]):int(L[4])])
    try:
        RNA_counts["U"] = RNA_counts.pop("T")  # renames T to U for RNA
    except KeyError:
        pass  # If there are no T's, pass

    # RNA Reaction formation
    calc_stoichiometry_df(RNA, RNA_counts)
    total = RNA.stoichiometry.sum()
    ids_fwd = list(RNA.id) + list(df.loc[["C00001"]].id.values)
    stoich_fwd = [RNA.loc["A"].stoichiometry + 2*total] +\
        list(RNA.stoichiometry)[1:] + [2*total]
    compound_fwd = {Compound(id).in_compartment(cell_compartment):
                    round(-1*n, decimals) for id, n in zip(ids_fwd,
                                                           stoich_fwd)}
    ids_rev = list(df.loc[["rna", "C00008", "C00009", "C00013"]].id.values)
    stoich_rev = [1, 2*total, 2*total, total]
    compound_rev = {Compound(id).in_compartment(cell_compartment):
                    round(n, decimals) for id, n in zip(ids_rev, stoich_rev)}
    rna_rxn = Reaction(Direction.Forward, {**compound_fwd, **compound_rev})
    rna_rxn_entry = {"id": "rna_met",
                     "name": "RNA",
                     "equation": str(rna_rxn),
                     "pathways": ["Biomass"]}
    return rna_rxn_entry


def return_biomass_rxn(df, biomass_name, cell_compartment):
    # Biomass reaction formation
    compound_fwd = {Compound(id).in_compartment(cell_compartment): -1
                    for id in list(df.loc[["dna",
                                           "rna", "protein"]].id.values)}
    compound_rev = {
        Compound(df.loc["biomass"].id).in_compartment(cell_compartment): 1}
    bio_rxn = Reaction(Direction.Forward, {**compound_fwd, **compound_rev})

    bio_rxn_entry = {"id": biomass_name,
                     "name": "Biomass",
                     "equation": str(bio_rxn),
                     "pathways": ["Biomass"]}
    return bio_rxn_entry


def return_bio_sink_rxn(df, cell_compartment):
    # Biomass sink reaction
    bio_sink = Reaction(Direction.Forward, {
        Compound(df.loc["biomass"].id).in_compartment(cell_compartment): -1})

    bio_sink_entry = {"id": "sink_biomass",
                      "name": "Biomass accumulation",
                      "equation": str(bio_sink),
                      "pathways": ["Biomass"]}
    return bio_sink_entry


def return_trna_rxns(df, cell_compartment):
    # tRNA charging reactions
    tRNA_rxns = []
    df_reac = pd.read_csv(
        resource_filename("psamm",
                          "external-data/biomass_reaction_descriptions.tsv"),
        sep="\t")
    for index, row in df_reac.iterrows():
        fwd, rev = row.equation.split(" <=> ")
        ids_fwd = fwd.split(" + ")
        ids_rev = rev.split(" + ")
        compound_fwd = {Compound(id).in_compartment(cell_compartment): -1
                        for id in list(df.loc[ids_fwd].id)}
        compound_rev = {Compound(id).in_compartment(cell_compartment): 1
                        for id in list(df.loc[ids_rev].id)}
        tRNA_rxn = Reaction(Direction.Both, {**compound_fwd, **compound_rev})
        tRNA_rxn_entry = {"id": row.id,
                          "name": row["name"],
                          "equation": str(tRNA_rxn),
                          "enzyme": row["enzymes"],
                          "pathways": ["Biomass"]}
        tRNA_rxns.append(tRNA_rxn_entry)
    return tRNA_rxns


class main(Command):
    """Generate a database of compounds and reactions"""

    @classmethod
    def init_parser(cls, parser):
        parser.add_argument("--genome", metavar="path",
                            help="[REQUIRED] Path to the genome "
                            "in fasta format")
        parser.add_argument("--proteome", metavar="path",
                            help="[REQUIRED] Path to the proteome "
                            "in fasta format")
        parser.add_argument("--gff", metavar="path",
                            help="[REQUIRED] Specify path to gff containing "
                            "transcript annotations. Only annotations "
                            "specified as \"CDS\" in the third column will be "
                            "used. Annotations must correspond to sequences in"
                            " the --genome file.")
        parser.add_argument("--model", metavar="path",
                            help="[REQUIRED] Path to the model file")
        parser.add_argument("--biomass", metavar="name", default="biomass",
                            help="Name to use for the biomass reaction, "
                            "default 'biomass'")
        parser.add_argument("--config", metavar="path",
                            help="If you are using custom compound names for "
                            "the biomass-related compounds (e.g. amino acids, "
                            "DNA bases), specify them via a 3-column config "
                            "file. Use --generate-config to create a template "
                            "file.")
        parser.add_argument("--generate-config", action="store_true",
                            help="Makes a template config file for --config. "
                            "Compound IDs added in the third column will "
                            "replace the default kegg IDs for the compound.")
        super(main, cls).init_parser(parser)

    def run(self):
        """Entry point for the biomass reaction generation script"""

        if not self._args.generate_config:
            if not self._args.genome:
                raise InputError("Please specify a path to the genome "
                                 "in fasta format with --genome")
            if not self._args.proteome:
                raise InputError("Specify path to proteome in fasta format "
                                 "with --proteome")
            if not self._args.gff:
                raise InputError("Specify path to gff containing transcript "
                                 "annotations with --gff.")
            if not self._args.model:
                raise InputError("Please specify the path to an existing "
                                 "model with --model")
            if os.path.isdir(self._args.model):
                if os.path.exists(os.path.join(self._args.model,
                                               "model.yaml")):
                    model_path = os.path.join(self._args.model, "model.yaml")
                else:
                    raise InputError("No model.yaml file found!")
            else:
                model_path = self._args.model

        # SETTING UP DATA
        pd.options.mode.chained_assignment = None
        df = load_compound_data()
        if self._args.generate_config:
            template = make_template(df)
            template.to_csv(sys.stdout, index=False)
            quit()

        faa = self._args.proteome
        fna = self._args.genome
        gff = self._args.gff
        model_dir = os.path.dirname(model_path)

        yaml_args = {"default_flow_style": False,
                     "sort_keys": False,
                     "encoding": "utf-8",
                     "allow_unicode": True,
                     "width": float("inf")}

        # Handling the config file
        if self._args.config:
            df = apply_config(df, self._args.config)

        # Adds missing compounds, updates model.yaml
        # Makes biomass compounds and reactions yamls
        logger.info("Checking for existing compounds in the model")
        check_missing_cpds(model_path, df, self._args.config, yaml_args)
        model_dict = load_model(model_path)
        cell_compartment = fix_cell_compartment(model_dict)

        logger.info("Updating {}".format(os.path.abspath(model_path)))
        update_model_yaml(model_path, model_dict, cell_compartment,
                          self._args.biomass, yaml_args)

        # Counting bases/amino acids and making reaction entries
        genome = {seq.name: seq.upper() for seq in SeqIO.parse(fna, "fasta")}
        proteome = {seq.name: seq.upper() for seq in SeqIO.parse(faa, "fasta")}

        logger.info("Calculating stoichiometry for DNA, RNA, and amino acids")
        dna_rxn_entry = count_DNA(genome, df, cell_compartment)
        prot_rxn_entry = count_Prot(proteome, df, cell_compartment)
        rna_rxn_entry = count_RNA(genome, gff, df, cell_compartment)

        # Making the reaction entries for other biomass rxns
        bio_rxn_entry = return_biomass_rxn(
            df, self._args.biomass, cell_compartment)
        bio_sink_entry = return_bio_sink_rxn(df, cell_compartment)
        tRNA_rxns = return_trna_rxns(df, cell_compartment)

        # GENERATING biomass_reactions.yaml
        logger.info("Generating new biomass reactions in "
                    "{}/biomass_reactions.yaml".format(
                        os.path.abspath(model_dir)))

        with open(os.path.join(model_dir, "biomass_reactions.yaml"), "w") as f:
            yaml.dump([dna_rxn_entry, rna_rxn_entry,
                       prot_rxn_entry, bio_rxn_entry,
                       bio_sink_entry] + tRNA_rxns,
                      f, **yaml_args)



#!/usr/bin/env python
# This file is part of PSAMM.
#
# PSAMM is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PSAMM is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PSAMM.  If not, see <http://www.gnu.org/licenses/>.
#
# Copyright 2015  Jon Lund Steffensen <jon_steffensen@uri.edu>
# Copyright 2020  Jing Wang <wjingsjtu@gmail.com>

import unittest
import tempfile
import shutil
import os

from psamm import manual_curation as curation
from psamm.datasource.native import ModelReader
from psamm.datasource.reaction import Compound


class TestCurator(unittest.TestCase):
    def setUp(self):
        self._model_dir = tempfile.mkdtemp()
        with open(os.path.join(self._model_dir, 'compound.tsv'), 'w') as f:
            f.write('id1\tid2\tp\n')
            f.write('%s\t%s\t%f\n' % ('A', 'A1', 1.0))
            f.write('%s\t%s\t%f\n' % ('B', 'B1', 1.0))
            f.write('%s\t%s\t%f\n' % ('C', 'C1', 0.03))
            f.write('%s\t%s\t%f\n' % ('D', 'D1', 0.00005))
        with open(os.path.join(self._model_dir, 'reaction.tsv'), 'w') as f:
            f.write('id1\tid2\tp\n')
            f.write('%s\t%s\t%f\n' % ('rxn1', 'rxn1', 1.0))
            f.write('%s\t%s\t%f\n' % ('rxn2', 'rxn2', 1.0))
            f.write('%s\t%s\t%f\n' % ('rxn3', 'rxn3', 0.03))
            f.write('%s\t%s\t%f\n' % ('rxn4', 'rxn4', 0.00005))
        self.curator = curation.Curator(
            os.path.join(self._model_dir, 'compound.tsv'),
            os.path.join(self._model_dir, 'reaction.tsv'),
            os.path.join(self._model_dir, 'curated_compound.tsv'),
            os.path.join(self._model_dir, 'curated_reaction.tsv')
        )

    def tearDown(self):
        shutil.rmtree(self._model_dir)

    def test_add_mapping(self):
        self.curator.add_mapping(('A', 'A1'), 'c', True)
        self.curator.add_mapping(('D', 'D1'), 'c', False)
        self.curator.add_mapping(('rxn1', 'rxn1'), 'r', True)
        self.curator.add_mapping(('rxn4', 'rxn4'), 'r', False)
        self.curator.save()

    def test_compound_checked(self):
        self.assertFalse(self.curator.compound_checked(('B', 'B1')))
        self.curator.add_mapping(('B', 'B1'), 'c', True)
        self.assertFalse(self.curator.compound_checked(('C', 'C1')))
        self.curator.save()
        curator = self.curator = curation.Curator(
            self.curator.compound_map_file,
            self.curator.reaction_map_file,
            self.curator.curated_compound_map_file,
            self.curator.curated_reaction_map_file
        )
        self.assertTrue(curator.compound_checked(('B', 'B1')))
        self.assertTrue(curator.compound_checked('B'))

    def test_compound_ignored(self):
        self.assertFalse(self.curator.compound_ignored('C'))
        self.curator.add_ignore('C', 'c')
        self.assertTrue(self.curator.compound_ignored('C'))
        self.curator.save()
        curator = self.curator = curation.Curator(
            self.curator.compound_map_file,
            self.curator.reaction_map_file,
            self.curator.curated_compound_map_file,
            self.curator.curated_reaction_map_file
        )
        self.assertTrue(curator.compound_ignored('C'))

    def test_reaction_checked(self):
        self.assertFalse(self.curator.reaction_checked(('rxn2', 'rxn2')))
        self.curator.add_mapping(('rxn2', 'rxn2'), 'r', False)
        self.assertFalse(self.curator.reaction_checked(('rxn3', 'rxn3')))
        self.curator.save()
        curator = self.curator = curation.Curator(
            self.curator.compound_map_file,
            self.curator.reaction_map_file,
            self.curator.curated_compound_map_file,
            self.curator.curated_reaction_map_file
        )
        self.assertTrue(curator.reaction_checked(('rxn2', 'rxn2')))

    def test_reaction_ignored(self):
        self.assertFalse(self.curator.reaction_ignored('rxn3'))
        self.curator.add_ignore('rxn3', 'r')
        self.assertTrue(self.curator.reaction_ignored('rxn3'))
        self.curator.save()
        curator = self.curator = curation.Curator(
            self.curator.compound_map_file,
            self.curator.reaction_map_file,
            self.curator.curated_compound_map_file,
            self.curator.curated_reaction_map_file
        )
        self.assertTrue(curator.reaction_ignored('rxn3'))


class TestSearch(unittest.TestCase):
    def setUp(self):
        self._model_dir = tempfile.mkdtemp()
        with open(os.path.join(self._model_dir, 'model.yaml'), 'w') as f:
            f.write('\n'.join([
                '---',
                'reactions:',
                '  - id: rxn_1',
                '    name: rxn_1',
                '    equation: A[e] => B[c]',
                '    genes: gene1 and gene2 or gene3',
                '  - id: rxn_2',
                '    name: rxn_2',
                '    equation: B[c] => C[e]',
                '    genes: gene5 or gene6',
                '  - id: rxn_3',
                '    name: rxn_3',
                '    equation: A[e] + (6) B[c] <=> (6) C[e] + (6) D[c]',
                '    genes: gene7',
                'compounds:',
                '  - id: A',
                '    name: compound_A',
                '    formula: C6H12O6',
                '    charge: 1',
                '  - id: B',
                '    name: compound_B',
                '    formula: O2',
                '    charge: 1',
                '    kegg: C00010',
                '  - id: C',
                '    name: compound_C',
                '    formula: CO2',
                '    charge: -1',
                '  - id: D',
                '    name: compound_D',
                '    formula: H2O',
            ]))
        self._model = ModelReader.reader_from_path(
            os.path.join(self._model_dir, 'model.yaml')).create_model()

    def tearDown(self):
        shutil.rmtree(self._model_dir)

    def test_search_compound(self):
        curation.search_compound(self._model, ['C', 'Z'])

    def test_search_reaction(self):
        cpds = curation.search_reaction(self._model, ['rxn_1', 'test'])
        self.assertListEqual(
            next(cpds),
            [Compound('A', 'e'), Compound('B', 'c')])

    def test_filter_search_term(self):
        self.assertEqual(curation.filter_search_term('Test-(tes)_t'),
                         'testtest')
